import re
import os
import json

from gimfu.t2listingh5 import t2listingh5
from gimfu.geo_common import quick_enthalpy

from t2data import t2generator
from t2data import t2data_parser, t2data
from t2incons import t2incon
from t2thermo import *
import numpy as np

KGSPERTD = 1000./24./60./60. # kg/s per ton/day
KGSPERTH = 1000./60./60. # kg/s per ton/hour

def normalname_match(name_wc, name):
    """ return true if the name_wc (wild cast * supported) match name """
    is_match = True
    for i in xrange(5):
        if name_wc[i] != '*':
            if name_wc[i] != name[i]:
                is_match = False
                break
    return is_match

def toughname_match(name_wc, name):
    """ return true if the name_wc (wild cast * supported) match name """
    from mulgrids import unfix_blockname
    c_name_wc = unfix_blockname(name_wc)
    c_name = unfix_blockname(name)

    is_match = True
    for i in xrange(5):
        if c_name_wc[i] != '*':
            if c_name_wc[i] != c_name[i]:
                is_match = False
                break
    return is_match

def read_gener_file(fname):
    f = t2data_parser(fname,'r')
    line = f.readline()
    if line[:5] != 'GENER':
        # get rid of first line 'GENER' and check
        print('!!!!!!!!!! ERROR !!!!!!!!!!')
        print('Error reading GENER file: ', fname)
    gs = t2data()
    gs.read_generators(f)
    f.close()
    return gs.generatorlist

def write_geners(geners, fname, silent=True):
    if not silent:
        print('Writing %s ...' % fname)
    tmpdat = t2data()
    outfile=t2data_parser(fname,'w')
    outfile.write('GENER\n')
    for g in geners:
        tmpdat.write_generator(g,outfile)
    outfile.close()

def write_gener_file(geners, filename, silent=True):
    if not silent:
        print('Writing %s ...' % filename)
    dat = t2data()
    for g in geners:
        dat.add_generator(g)
    outfile=t2data_parser(filename,'w')
    dat.write_generators(outfile)
    outfile.close()

def write_gener_if(geners, is_true, filename):
    """ This will go through a list of t2generators, and only writes to file is
    individual generator is qualified by is_true.  is_true is a callable,
    accepting a single t2generator object, returns True or False.
    """
    dat = t2data()
    for g in geners:
        if is_true(g):
            dat.add_generator(g)
    outfile=t2data_parser(filename,'w')
    dat.write_generators(outfile)
    outfile.close()

def select_geners_byname(names, geners, warning=False):
    selected = []
    for gn in names:
        gre = re.compile(gn)
        gs = [g for g in geners if gre.match(g.name)]
        if warning and len(gs) == 0:
            raise Exception("No gener(s) match %s" % gn)
        selected += gs
    return selected

def well_blocks(geo, well_name, divisions=1, elevation=False,
    deviations=False, qtree=None, extend=False):
    blocks = geo.well_values(well_name,
                             geo.block_name_list,
                             divisions=divisions,
                             elevation=elevation,
                             deviations=deviations,
                             qtree=qtree,
                             extend=extend)
    if blocks is None:
        raise Exception('Well %s not found in geometry %s' % (well_name, geo.filename))
    return blocks[1]

def modify_wellname(mod,original):
    """ modifying well name according to rules: any '*' characters in mod
    will be keep original name, otherwise overwritten by mod. """
    newname = []
    for i in range(5):
        if mod[i] == '*':
            newname.append(original[i])
        else:
            newname.append(mod[i])
    return "".join(newname)

def imak_group_by_ratios(ratios, total, geners,
                         overflow_unit_ratio=0.1, overflow_times=None,
                         warning=False):
    """
    overflow_unit_ratio between 0.0 and 1.0
    if overflow_times is integer, repeat base geners n times with the ratio
    """
    all_geners = []
    base_geners = []
    for wn, ratio in ratios:
        wgs = select_geners_byname([wn], geners, warning=True)
        for g in wgs:
            g.gx = total * ratio / float(len(wgs))
            base_geners.append(g)
            all_geners.append(g)
    from copy import deepcopy
    if isinstance(overflow_times, int):
        for i in range(overflow_times):
            for g in base_geners:
                gg = deepcopy(g)
                gg.gx = g.gx * overflow_unit_ratio
                all_geners.append(gg)
    return all_geners

def make_IMAK_wells(incon, addp, pi, temp, newwell_labels, blocks, rates,
                    new_total_rate):
    """  returns a list of generators. if newwell_labels contains '*' the '*'s
    will be replaced by an incrementing integer.
    """

    total_rate = sum(rates)
    dat = t2data()
    for i in range(len(blocks)):
        gx = rates[i] / total_rate * new_total_rate
        ex = quick_enthalpy(temp,'liq')
        hg = incon[b][0] + addp
        fg = pi

        w_name = modify_wellname(newwell_labels,'%05d'%(i+1))
        g=t2generator(name=w_name,block=blocks[i],type='IMAK',gx=gx,ex=ex,hg=hg,fg=fg)
        dat.add_generator(g)
    return dat.generatorlist

def make_IMAK_by_blocks(total_rate, blocks_ratios, incon, addp, pi, temp,
                        geo,
                        naming='II***',
                        overflow_unit_ratio=0.1, overflow_times=None,
                        warning=False):
    """ if incon is None, use rho*g*h + addp as IMAK pressure, otherwise use
    incon of the block to add p

    Usually block_ratios is a list of tuples of (block, ratio).  But it is
    possible to have tuples of three (block, ratio, naming), where the naming
    will overwrite the routine level naming.
    """

    rho = cowat(temp, sat(temp)+1.0e5)[0]

    all_geners = []
    base_geners = []
    for br in blocks_ratios:
        if len(br) == 2:
            b, ratio = br
            final_naming = naming
        elif len(br) == 3:
            b, ratio, final_naming = br

        if isinstance(addp, dict):
            _addp = addp[b]
        else:
            _addp = addp


        if incon is None:
            col = geo.column[geo.column_name(b)]
            lay = geo.layer[geo.layer_name(b)]
            pbottom = rho * 9.81 * (col.surface - geo.block_centre(lay, col)[2])
            pbottom += _addp
        else:
            pbottom = incon[b][0] + _addp

        # print('# making GENERs:')
        g = t2generator(name=modify_wellname(final_naming,b),
                        block=b,
                        type='IMAK',
                        gx=total_rate * ratio,
                        ex=quick_enthalpy(temp,'liq'),
                        hg=pbottom,
                        fg=pi,
                        )
        base_geners.append(g)
        all_geners.append(g)

    from copy import deepcopy
    if isinstance(overflow_times, int):
        for i in range(overflow_times):
            for g in base_geners:
                gg = deepcopy(g)
                gg.gx = g.gx * overflow_unit_ratio
                all_geners.append(gg)
    return all_geners




def make_IMAK_by_wellnames(total_rate, wellnames_ratios, incon, addp, pi, temp,
                           feedzone_fname, geo,
                           naming='II***',
                           overflow_unit_ratio=0.1, overflow_times=None,
                           warning=False):
    """ if incon is None, use rho*g*h + addp as IMAK pressure,
    otherwise use incon of the block to add p.

    For each inj well to have its own WHP, addp can be specified as a dict.  The
    keys should be wellnames, same as used for ratios and feedzones.
    """

    rho = cowat(temp, sat(temp)+1.0e5)[0]


    with open(feedzone_fname, 'r') as f:
        data_feeds = json.load(f)

    all_geners = []
    base_geners = []
    for wn,ratio in wellnames_ratios:
        # print('# checking %s feedzones' % wn,)
        bs = well_blocks(geo, wn)
        feed_blocks = []
        for f in data_feeds[wn]:
            for b in bs:
                blay = geo.layer[geo.layer_name(b)]
                btop, bbot = blay.top, blay.bottom
                ftop, fbot = f['elev_top'], f['elev_bot']
                if (ftop > bbot > fbot) or (ftop > btop > fbot) or (btop >= ftop and bbot <= fbot):
                    # print(b,)
                    if b not in feed_blocks:
                        feed_blocks.append(b)
        if isinstance(addp, dict):
            _addp = addp[wn]
        else:
            _addp = addp
        for b in feed_blocks:
            if incon is None:
                col = geo.column[geo.column_name(b)]
                lay = geo.layer[geo.layer_name(b)]
                pbottom = rho * 9.81 * (col.surface - geo.block_centre(lay, col)[2])
                pbottom += _addp
            else:
                pbottom = incon[b][0] + _addp
            g = t2generator(name=modify_wellname(naming,wn),
                            block=b,
                            type='IMAK',
                            gx=total_rate * ratio / float(len(feed_blocks)),
                            ex=quick_enthalpy(temp,'liq'),
                            hg=pbottom,
                            fg=pi,
                            )
            base_geners.append(g)
            all_geners.append(g)

    from copy import deepcopy
    if isinstance(overflow_times, int):
        for i in range(overflow_times):
            for g in base_geners:
                gg = deepcopy(g)
                gg.gx = g.gx * overflow_unit_ratio
                all_geners.append(gg)
    return all_geners


def make_IMAK_by_well(total_rate, wellnames_ratios, incon, addp, pi, temp,
                           feed_block_ratios, geo,
                           naming='II***',
                           overflow_unit_ratio=0.1, overflow_times=None,
                           warning=False):
    """ if incon is None, use rho*g*h + addp as IMAK pressure,
    otherwise use incon of the block to add p.

    feed_block_ratios is a dict of wellname: ["block": ratio, ...]

    For each inj well to have its own WHP, addp can be specified as a dict.  The
    keys should be wellnames, same as used for ratios and feedzones.
    """
    rho = cowat(temp, sat(temp)+1.0e5)[0]

    all_geners = []
    base_geners = []
    for wn,w_ratio in wellnames_ratios:
        if isinstance(addp, dict):
            _addp = addp[wn]
        else:
            _addp = addp
        for b,b_ratio in feed_block_ratios[wn].items():
            if incon is None:
                col = geo.column[geo.column_name(b)]
                lay = geo.layer[geo.layer_name(b)]
                pbottom = rho * 9.81 * (col.surface - geo.block_centre(lay, col)[2])
                pbottom += _addp
            else:
                pbottom = incon[b][0] + _addp
            g = t2generator(name=modify_wellname(naming,wn),
                            block=b,
                            type='IMAK',
                            gx=total_rate * w_ratio * b_ratio,
                            ex=quick_enthalpy(temp,'liq'),
                            hg=pbottom,
                            fg=pi,
                            )
            base_geners.append(g)
            all_geners.append(g)

    from copy import deepcopy
    if isinstance(overflow_times, int):
        for i in range(overflow_times):
            for g in base_geners:
                gg = deepcopy(g)
                gg.gx = g.gx * overflow_unit_ratio
                all_geners.append(gg)
    return all_geners


def make_injection_IMAK_group(geo, group_naming, rate, temp, wells, whp=None,
                              incon=None, pi=None, is_condensate=False, max_whp=None,
                              feedzone_filename=None, warning=True,
                              overflow_unit_ratio=1.0, overflow_times=0):

    """ make a group of IMAK geners that cap total to the specified rate.  it
    can be specified by either a list of wells and/or blocks.  Returns a list of
    T2 generators.

    If incon is None, injection pressure = rho * g * depth + whp. If incon is a
    t2incon object, whp is 'additional pressure', added onto the block's
    pressure from t2incon, injection pressure = whp + block_pressure.

    If max_whp is specified, it will act as an additional cap for all specified
    WHPs.

    To specify wells, use a list of tuples:
        (well name, block, ratio, whp, pi)
    Here the second element is set to None, blocks of the well will be worked
    out from the feedzone information.  If block is a valid string, then
    generator will be created using well name.  If specifying by block, well
    can be left as None if the gener is to be named as the block name.  eg.
        ('WK123',    None, 0.3, 10.0*e5, 1.0e-11), # WK123 may have several blocks,
        ('IA124', 'asd21', 0.7, 11.0*e5, 1.0e-11), # gener.name = 'IA124'
        (None,    'asd21', 0.7, 11.0*e5, 1.0e-11), # gener.name = 'asd21'

    whp and pi can be specified at the well/block level.  If the value is None
    in individual well/block, then it will use the default single value.
    """
    if is_condensate:
        pi_factor = -1.0
    else:
        pi_factor = 1.0

    # print('making %s %.1f kg/s ...' % (group_naming, rate))
    gs = []
    for wname, block, ratio, _whp, _pi in wells:
        if _whp is None: _whp = whp
        if max_whp is not None:
            _whp = min(_whp, max_whp)
        if _pi is None: _pi = pi
        if wname is None: wname = block
        if block is None:
            # by wells
            gs +=make_IMAK_by_wellnames(rate, [(wname, ratio)], incon=incon,
                                        addp={wname: _whp}, pi=_pi*pi_factor,
                                        temp=temp,
                                        feedzone_fname=feedzone_filename,
                                        geo=geo, naming=group_naming,
                                        overflow_unit_ratio=0.0,
                                        overflow_times=0,
                                        warning=warning)
        else:
            # by block
            gs += make_IMAK_by_blocks(rate, [(block, ratio, wname)], incon=incon,
                                      addp=_whp, pi=_pi*pi_factor, temp=temp,
                                      geo=geo, naming=group_naming,
                                      overflow_unit_ratio=0.0,
                                      overflow_times=0,
                                      warning=warning)
    for gener in gs:
        gener.name = modify_wellname(group_naming, gener.name)

    from copy import deepcopy
    gs_overflow = []
    if overflow_times > 0:
        for g in gs:
            gg = deepcopy(g)
            gg.gx = g.gx * overflow_unit_ratio
            gs_overflow.append(gg)

    return gs + gs_overflow * overflow_times


def test_make_injection_IMAK_group():
    # test 1
    naming = 'IA7'
    naming2 = naming + (5-len(naming)) * '*'
    rate = 5. *1000.*KGSPERTD
    temp = 108.0
    whp = 10.0e5

    bs = ['mar48', 'mar49', 'nar48', 'nar49', ]
    grp_def = [
        ('mar48', 'mar48', 0.25, None, None),
        ('mar49', 'mar49', 0.25, None, None),
        ('nar48', 'nar48', 0.25, None, None),
        ('nar49', 'nar49', 0.25, None, None),
    ]

    write_geners(inj_sgw_makeup(naming, rate, temp, bs, whp),
                 'inj_test1_A.geners')
    write_geners(make_injection_IMAK_group(geo, naming2, rate, temp, grp_def,
                                           whp=whp, incon=None, pi=INJECTIVITY,
                                           is_condensate=False,
                                           feedzone_filename=feed_zone_file,
                                           warning=True),
                 'inj_test1_B.geners')

    # test 2
    naming = 'IDK'
    naming2 = naming + (5-len(naming)) * '*'
    temp = 40.0
    whp = 12.0*1.0e5

    grp_def = [
        ('WK407',    None, 1.00, None, None),
    ]

    write_geners(inj_cnd_407('ID', rate, temp)[0],
                 'inj_test2_A.geners')
    write_geners(make_injection_IMAK_group(geo, naming2, rate, temp, grp_def,
                                           whp=whp, incon=None, pi=INJECTIVITY,
                                           is_condensate=True,
                                           feedzone_filename=feed_zone_file,
                                           warning=True) * 6,
                 'inj_test2_B.geners')

    # test 3
    naming = 'IC'
    naming2 = naming + (5-len(naming)) * '*'
    temp = 40.0
    whp = 12.0*1.0e5
    grp_def = [
        #  wname,   block, ratio,        whp,   pi
        ('WK301',    None,  0.05, 27.0*1.0e5, None),
        ('WK304',    None,  0.02, 20.0*1.0e5, None),
        ('WK308',    None,  0.25, 15.0*1.0e5, None),
        ('WK309',    None,  0.15, 16.0*1.0e5, None),
        ('WK310',    None,  0.03, 20.0*1.0e5, None),
        ('WK317',    None,  0.15, 38.0*1.0e5, None),
        ('WK318',    None,  0.15, 20.0*1.0e5, None),
        ('WK321',    None,  0.20, 36.0*1.0e5, None),
        ]

    write_geners(inj_cnd_otupu(naming, rate, temp, max_whp=None)[0],
                 'inj_test3_A.geners')
    write_geners(make_injection_IMAK_group(geo, naming2, rate, temp, grp_def,
                                           whp=whp, incon=None, pi=INJECTIVITY,
                                           is_condensate=True,
                                           feedzone_filename=feed_zone_file,
                                           warning=True,
                                           overflow_unit_ratio=0.5,
                                           overflow_times=1),
                 'inj_test3_B.geners')

    # test 4
    naming = 'IK'
    naming2 = naming + (5-len(naming)) * '*'
    temp = 108.0
    whp = 12.0*1.0e5
    grp_def = [
        #  wname,   block, ratio,        whp,   pi
        ('WK401',    None,  0.10, 13.2*1.0e5, None),
        ('WK403',    None,  0.25, 19.5*1.0e5, None),
        ('WK404',    None,  0.15, 45.4*1.0e5, None),
        ('WK408',    None,  0.15, 53.0*1.0e5, None),
        ('WK409',    None,  0.10, 34.2*1.0e5, None),
        ('WK410',    None,  0.25, 29.5*1.0e5, None),
    ]

    write_geners(inj_sgw_karapiti(naming, rate, temp, max_whp=None)[0],
                 'inj_test4_A.geners')
    write_geners(make_injection_IMAK_group(geo, naming2, rate, temp, grp_def,
                                           whp=whp, incon=None, pi=INJECTIVITY,
                                           is_condensate=False,
                                           feedzone_filename=feed_zone_file,
                                           warning=True,
                                           overflow_unit_ratio=0.5,
                                           overflow_times=0),
                 'inj_test4_B.geners')

    # test 5 (mixed)
    naming = 'IK'
    naming2 = naming + (5-len(naming)) * '*'
    temp = 108.0
    whp = 12.0*1.0e5
    grp_def = [
        ('WK407',    None, 0.20, 12.e5, None),
        ('mar48', 'mar48', 0.20, 10.e5, None),
        ('mar49', 'mar49', 0.20, 10.e5, None),
        ('nar48', 'nar48', 0.20, 10.e5, None),
        ('nar49', 'nar49', 0.20, 10.e5, None),
    ]

    write_geners(make_injection_IMAK_group(geo, naming2, rate, temp, grp_def,
                                           whp=whp, incon=None, pi=INJECTIVITY,
                                           is_condensate=False,
                                           feedzone_filename=feed_zone_file,
                                           warning=True,
                                           overflow_unit_ratio=0.5,
                                           overflow_times=0),
                 'inj_test5.geners')


def block_depth(geo, block_name):
    """ get depth of a model block (from surface)

    If block_name is a list of names, then a list of depths will be returened.
    """
    if not isinstance(block_name, list):
        bs = [block_name]
    else:
        bs = block_name
    depths = []
    for b in bs:
        c = geo.column_name(b)
        lay = geo.layer[geo.layer_name(b)]
        try:
            depths.append(geo.column[c].surface - geo.block_centre(lay, geo.column[c])[2])
        except KeyError:
            # happends when column does not exist (special atm)
            depths.append(geo.layerlist[0].bottom - lay.centre)
    if not isinstance(block_name, list):
        return depths[0]
    else:
        return depth


def liq_inj_pressure(t, depth, pump):
    """ Returns pressure at depth by adding over-/pump- pressure to rho*g*h.

    temp in degc, depth in m, pump pressure in pa """
    import t2thermo
    # liq water density does not vary too much with pressure
    # hence assume saturation pressure + 10 bar
    ADD_SAT_P = 10.0e5
    g = 9.81
    d, u = t2thermo.cowat(t, p=t2thermo.sat(t)+ADD_SAT_P)
    return d * g * depth + pump

class DeliverabilityCurvesSteam(object):
    """ a simplified, depth-independent curve used for steam wells
    """
    def __init__(self, pcutoff=2.5e5, phuge=300.e5):
        # all in (Pa)
        self.phuge = phuge
        self.pcutoff = pcutoff

    def get_pcutoff_table(self):
        w_enth = [0.0000000e+00, 2.0000000e+06, 2.1000000e+06, 2.8000000e+06]
        w_pres = [   self.phuge,    self.phuge,  self.pcutoff,  self.pcutoff]
        return w_enth, w_pres

    def __repr__(self):
        return 'steam-cutoff-%.2f-bar' % (self.pcutoff/1.e5)

class DeliverabilityCurves(object):
    """ Pcutoff vs. Enthalpy curve set for GENER types such as DELG and DMAK.

    The set includes a Pcutoff vs Enthalpies for a set of Depths.

    hard_cutoff_enth - (J/kg) is the *hard* cut-off enthalpy, the table was
    overwrittend below this point to use the p_huge value.  If None, table's
    enthalpies will not be altered.

    p_huge - (Pa) is the huge value used for hard cutoff
    """
    def __init__(self, filename=None):
        self.depths = None # a list of depths (m)
        self.enths = None  # a list of enthalpies (J/kg)
        self.table = None  # a numpy () array of Pcutoffs (Pa) (n_depths, n_enths)
        self.hard_cutoff_enth = None # (J/kg)
        self.p_huge = 500.0e5 # (Pa)
        self.curve_source = 'cfg'
        if filename is not None:
            self.load_from_cfg(filename)

    def load_from_cfg(self, filename):
        import config
        if isinstance(filename, config.config):
            cfg = filename
        else:
            self.curve_source = filename
            cfg = config.config(filename)
        # unfortunately old curve.cfg format uses kJ/kg and bar here
        # load the tables and depths values
        self.depths = [float(a) for a in cfg.get_list('LookupDepths')]
        self.enths = [float(a)*1000.0 for a in cfg.get_list('PcutoffEnthalpyTable')[0].split()]
        table_raw = cfg.get_list('PcutoffEnthalpyTable')
        self.table = np.array([[float(a)*1.0e5 for a in table_raw[i].split()] for i in range(1,len(table_raw))])
        # array of shape (len(depths), len(enths))

        if cfg.check_optional('CutoffEnthalpy'):
            self.hard_cutoff_enth = float(cfg.get_value('CutoffEnthalpy').strip()) * 1000.0
        if cfg.check_optional('HugePressure'):
            self.p_huge = float(cfg.get_value('HugePressure').strip()) * 1.0e5

    def get_pcutoff_table(self, depth):
        """ get a "Pcutoff vs. Enthalpy" curve for a given depth

        returns (enthalpy list, pressure list), can be used in DELG/DMAK etc.
        """
        SMALL_ENTH = 10.0 * 1000.0 # small increment of enthalpy for creating the cutoff region
        if (depth < self.depths[0]) or (depth > self.depths[-1]):
            raise Exception('!!! Depth of %f out of DeliverabilityCurves range: (%f,%f)' % (depth, self.depths[0], self.depths[-1]))

        pressures = [np.interp(depth, self.depths, self.table[:,e]) for e in range(len(self.enths))]

        # start table to cover zero
        w_enth, w_pres = [0.0], [self.p_huge]
        if self.hard_cutoff_enth is not None:
            # pad if hard cutoff
            p_enth_add10 = np.interp(self.hard_cutoff_enth + SMALL_ENTH , self.enths, pressures)
            w_enth += [self.hard_cutoff_enth, self.hard_cutoff_enth+SMALL_ENTH]
            w_pres += [self.p_huge, p_enth_add10]
        else:
            # otherwise cutoff at SMALL_ENTH
            w_enth += [self.enths[0] + SMALL_ENTH]
            w_pres += [self.p_huge]

        # form the table to be inserted into generator
        ii = [i for i in range(len(self.enths)) if self.enths[i] > w_enth[-1]]
        w_enth += [self.enths[i] for i in ii]
        w_pres += [pressures[i] for i in ii]

        # add one last entry if the table may not cover all enthalpy situations
        if w_enth[-1] < 2800.0e3:
            w_enth.append(2800.0e3)
            w_pres.append(w_pres[-1])

        return w_enth, w_pres

    def __repr__(self):
        return self.curve_source


def get_delg_pis(geners, pr_filename, sav_filename, init_massflow=None,
                 tmpdir='tmp_get_pi_delg', aut2='AUTOUGH2_7_0-10',
                 not_flowing_pi=None, pr_listing=None):
    """ Using AUTOUGH2's .autogeners to get PIs for DELG/DMAK/DMAT generators.

    TOUGH2 input file name expected here instead of t2data object because we
    want to open a new instance of t2data for temporary use, just to avoid
    potential confusion with whatever that is passed in.

    Output geners will have .gx overwritten by PI calculated by AUT2 in all
    cases.  But other values such as .type .ex .hg .fg will be kept as the
    original geners.

    This subroutine also supports metadata, if the original t2generator has
    additional property of .meta then it will be inherited by the returned
    generators.  It doesn't matter what object it is.

    If not_flowing_pi is set, it will be used to replace the 0.0 PI caused by
    generators that are not flowing (possibly due to pressure too low).

    NOTE if the well is not able to flow, PI will be 0.0 (from AUT2)
    """
    # only works with these, others may have incompatable gener properties
    for g in geners:
        if g.type not in ['DELG', 'DMAK', 'DMAT']:
            raise Exception("get_delg_pis() does not support gener '%s' with type: '%s'" % (str(g), g.type))

    if init_massflow is None:
        # use rates with original .hg
        init_rates = [-abs(g.hg) for g in geners]
    elif isinstance(init_massflow, list):
        if len(init_massflow) != len(geners):
            raise Exception("get_delg_pis() init_massflow does not have correct length!")
        init_rates = [-abs(f) for f in init_massflow]
    elif isinstance(init_massflow, float):
        init_rates = [-abs(init_massflow)] * len(geners)

    orig_meta = {(g.block, g.name): g.meta for g in geners if hasattr(g, 'meta')}
    orig_types = [g.type for g in geners]
    orig_hgs = [g.hg for g in geners]
    orig_fgs = [g.fg for g in geners]
    tmp = t2data(pr_filename)
    tmp.clear_generators()
    tmp.parameter['max_timesteps'] = 1
    tmp.parameter['max_timestep'] = 1.0 # sec
    tmp.parameter['print_level'] = 1 # supress CCCCC table dump
    for g, ir in zip(geners, init_rates):
        g.type = 'DELG'
        g.hg = ir
        tmp.add_generator(g)
    sav = t2incon(sav_filename)

    if not os.path.isdir(tmpdir):
        os.mkdir(tmpdir)
    original_dir = os.getcwd()
    os.chdir(tmpdir)

    # ---
    tmp.write('tmp.dat')
    sav.write('tmp.incon', reset=True)
    tmp.run(save_filename='tmp.save',
            incon_filename='tmp.incon',
            simulator=aut2,
            silent=True)
    auto_gs = read_gener_file('tmp.autogeners')
    lst = t2listingh5('tmp.h5')
    lst.last()
    if pr_listing:
        pr_listing.last()
    for g,t,hg,fg in zip(auto_gs, orig_types, orig_hgs, orig_fgs):
        if g.gx == 0.0:
            # find out how much too low pressure is
            enth = lst.generation[(g.block, g.name)]['Enthalpy']

            if pr_listing:
                for block, name in pr_listing.generation.row_name:
                    if g.block == block:
                        enth = pr_listing.generation[(g.block, name)]['Enthalpy']

            pres = lst.element[g.block]['Pressure']
            p_cutoff = np.interp(enth, g.time, g.rate)

            if not_flowing_pi is not None:
                print('*** %s:%s possibly pressure too low to flow, reset PI to %10.4e' % (
                    g.block, g.name, not_flowing_pi))
                g.gx = not_flowing_pi
            else:
                print('*** %s:%s possibly pressure too low to flow, PI is ZERO' % (
                    g.block, g.name))

            print('                ' \
                  '%10.5f < %10.5f bar at %10.5f kJ/kg' % (
                  pres / 1.e5, p_cutoff / 1.e5, enth / 1.e3))
        g.type = t
        g.hg = hg
        g.fg = fg
        if (g.block, g.name) in orig_meta:
            g.meta = orig_meta[(g.block, g.name)]

    os.chdir(original_dir)
    return auto_gs


def get_pi_delg(geners, pr_filename, sav_filename, max_steam=20.0):
    """ Using AUTOUGH2's .autogeners to get PIs for DELG/DMAK generators.

    TOUGH2 input file name expected here instead of t2data object because we
    want to open a new instance of t2data for temporary use, just to avoid
    potential confusing with whatever that is passed in.
    """

    orig_types = [g.type for g in geners]
    orig_fgs = [g.fg for g in geners]
    tmp = t2data(pr_filename)
    tmp.clear_generators()
    tmp.parameter['max_timesteps'] = 1
    for g in geners:
        # g.type = 'DELG'
        tmp.add_generator(g)
    sav = t2incon(sav_filename)

    TMPDIR = 'tmp_get_pi_delg'
    if not os.path.isdir(TMPDIR):
        os.mkdir(TMPDIR)
    # for f in copy_files:
    #     shutil.copy2(f, TMPDIR)
    original_dir = os.getcwd()
    os.chdir(TMPDIR)

    # ---
    tmp.write('tmp.dat')
    sav.write('tmp.incon', reset=True)
    tmp.run(save_filename='tmp.save',
            incon_filename='tmp.incon',
            simulator=AUT2,
            silent=True)
    auto_gs = read_gener_file('tmp.autogeners')
    lst = None
    if max_steam is None:
        lst = t2listing('tmp.listing')
    for g,t,fg in zip(auto_gs, orig_types, orig_fgs):
        if g.gx == 0.0:
            print('*** %s:%s possibly pressure too low to flow, reset PI' % (g.block, g.name))
            g.gx = 1.0e-11
        if lst is not None:
            max_steam = lst.generation[(g.block, g.name)]['Steam sepa.']
        else:
            pass
        g.type = t
        g.hg = max_steam
        g.fg = fg

    os.chdir(original_dir)
    return auto_gs


def get_max_steam_delg(geners, pr_filename, sav_filename, max_steam=20.0,
                       large_pi=1.0e-10):
    """ use steam separation from listing gener tables to cap DMAK/DELGs

    Combined with a reasonably large PI, this can 'throttle' an actually more
    productive DMAK at a lower initial rate.

    If large_pi is set to None, then the original PI will be preserved.  Be
    careful this will mean the well's initial flowrate to be determined by both
    PI as well as steam cap!

    TOUGH2 input file name expected here instead of t2data object because we
    want to open a new instance of t2data for temporary use, just to avoid
    potential confusing with whatever that is passed in.
    """
    orig_types = [g.type for g in geners]
    orig_fgs = [g.fg for g in geners]
    tmp = t2data(pr_filename)
    tmp.clear_generators()
    tmp.parameter['max_timesteps'] = 1
    for g in geners:
        # g.type = 'DELG'
        tmp.add_generator(g)
    sav = t2incon(sav_filename)

    TMPDIR = 'tmp_get_max_steam_delg'
    if not os.path.isdir(TMPDIR):
        os.mkdir(TMPDIR)
    # for f in copy_files:
    #     shutil.copy2(f, TMPDIR)
    original_dir = os.getcwd()
    os.chdir(TMPDIR)

    # ---
    tmp.write('tmp.dat')
    sav.write('tmp.incon', reset=True)
    tmp.run(save_filename='tmp.save',
            incon_filename='tmp.incon',
            simulator=AUT2,
            silent=True)
    auto_gs = read_gener_file('tmp.autogeners')
    lst = t2listing('tmp.listing')
    for g,t,fg in zip(auto_gs, orig_types, orig_fgs):
        g.gx = large_pi
        steam = lst.generation[(g.block, g.name)]['Steam sepa.']
        if steam == 0.0:
            print('*** %s:%s possibly pressure too low to flow, use default steam max' % (g.block, g.name))
            steam = max_steam
        g.type = t
        g.hg = steam
        g.fg = fg

    os.chdir(original_dir)
    return auto_gs











from mulgrids import *
from t2data import *



def normalname_match(name_wc, name):
    """ return true if the name_wc (wild cast * supported) match name """
    is_match = True
    for i in range(5):
        if name_wc[i] != '*':
            if name_wc[i] != name[i]:
                is_match = False
                break
    return is_match

def toughname_match(name_wc, name):
    """ return true if the name_wc (wild cast * supported) match name """
    from mulgrids import unfix_blockname
    c_name_wc = unfix_blockname(name_wc)
    c_name = unfix_blockname(name)

    is_match = True
    for i in xrange(5):
        if c_name_wc[i] != '*':
            if c_name_wc[i] != c_name[i]:
                is_match = False
                break
    return is_match

# GENER purpose check

def is_upflow(gener,geo,convention='*** 1'):
    b = gener.block
    w = gener.name
    checks = [
        normalname_match(convention,w),
        geo.layer[geo.layer_name(b)].name == geo.layerlist[-1].name,
        gener.type == 'MASS'
        ]
    return all(checks)

def is_heat(gener,geo,convention='*** 2'):
    b = gener.block
    w = gener.name
    checks = [
        normalname_match(convention,w),
        geo.layer[geo.layer_name(b)].name == geo.layerlist[-1].name,
        gener.type == 'HEAT'
        ]
    return all(checks)

def is_upflow_rech(gener,geo,convention='*** 3'):
    b = gener.block
    w = gener.name
    checks = [
        normalname_match(convention,w),
        geo.layer[geo.layer_name(b)].name == geo.layerlist[-1].name,
        gener.type == 'RECH'
        ]
    return all(checks)

def is_side_rech(gener,geo,convention='*****'):
    b = gener.block
    w = gener.name
    checks = [
        normalname_match(convention,w),
        b == w,
        gener.type == 'RECH'
        ]
    return all(checks)

def is_rain(gener,geo,convention='*****'):
    b = gener.block
    w = gener.name
    checks = [
        normalname_match(convention,w),
        geo.layer_name(b) == geo.layerlist[geo.num_layers - geo.column[geo.column_name(b)].num_layers].name,
        gener.type == 'MASS'
        ]
    return all(checks)

def is_spring(gener,geo,convention='SP***'):
    b = gener.block
    w = gener.name
    checks = [
        normalname_match(convention,w),
        gener.type in ['DELG','MASS'],
        ]
    return all(checks)

def check_geners(geners, checks):
    """ geners, a list of t2generators.  checks, a list of functions that
    accepts generator as argument and return True/False, names of functions
    will be used as dict key when returning counts of each function when
    returning True.  A list of partial functions can be in place the list of
    functions. """

    try:
        fnames = [f.__name__ for f in checks]
    except AttributeError:
        fnames = [pf.func.__name__ for pf in checks]

    # set all counter to zero, dict with func names as key
    count = dict([(n,0) for n in fnames])
    count['UNKNOWN'] = 0

    for i,g in enumerate(geners):
        match = []
        for f,t in zip(checks,fnames):
            if f(g):
                count[t] += 1
                match.append(t)
        if len(match) > 1:
            print('WARNING: %s match types: %s' % (g.name,','.join(match)))
        if len(match) == 0:
            count['UNKNOWN'] += 1
            print('WARNING: %s doe not match any type.' % g.name)

    return count

def wai_area_polygon():
    pline = [
        np.array([ 2768202.24117734 , 6270490.54421648]),
        np.array([ 2776545.06074986 , 6276563.022725  ]),
        np.array([ 2777310.95894013 , 6277137.4463677 ]),
        np.array([ 2778049.5036236  , 6277766.57702399]),
        np.array([ 2780784.85430311 , 6280173.68562197]),
        np.array([ 2781988.4086021  , 6281377.23992095]),
        np.array([ 2783629.61900981 , 6294807.81175738]),
        np.array([ 2772524.09525097 , 6296312.25463111]),
        np.array([ 2764454.8107464  , 6291662.15847593]),
        np.array([ 2762567.41877753 , 6281213.11888018]),
        np.array([ 2764865.11334833 , 6272842.94580086]),
        np.array([ 2767436.34298707 , 6270955.553832  ])]
    return pline

def ro_area_polygon():
    pline = [
        np.array([ 2783739.03303699 , 6294671.0442234 ]),
        np.array([ 2781742.22704094 , 6281623.42148211]),
        np.array([ 2785216.12240393 , 6279462.49444529]),
        np.array([ 2787267.63541357 , 6279626.61548606]),
        np.array([ 2788416.48269896 , 6279134.25236375]),
        np.array([ 2795911.34356084 , 6279927.50406081]),
        np.array([ 2790440.64220181 , 6292263.93562543]),
        np.array([ 2784586.99174764 , 6294534.27668943])]
    return pline

def th_area_polygon():
    pline = [
        np.array([ 2788334.42217858 , 6279298.37340452]),
        np.array([ 2795938.69706763 , 6280118.97860838]),
        np.array([ 2797579.90747534 , 6272131.75462419]),
        np.array([ 2794188.07263274 , 6265129.25688463]),
        np.array([ 2789784.15803872 , 6262776.85530024]),
        np.array([ 2778842.75532066 , 6260479.16072945]),
        np.array([ 2771566.72251314 , 6264992.48935065]),
        np.array([ 2768421.0692317  , 6269970.82758737]),
        np.array([ 2778131.56414398 , 6277848.63754438]),
        np.array([ 2781632.81301376 , 6280966.93731903]),
        np.array([ 2784942.58733598 , 6279407.7874317 ]),
        np.array([ 2786747.91878446 , 6279653.96899286]),
        np.array([ 2788115.59412422 , 6279353.08041811])]
    return pline

if __name__ == '__main__':

    geo = mulgrid('gwai41458_09.dat')
    nsdat = t2data('wai41458ns_464.dat')
    prdat = t2data('wai41458pr_464.dat')

    # check these possibilities
    from functools import partial
    g_types = [
    partial(is_rain,geo=geo,convention='*** 0'),
    partial(is_heat,geo=geo,convention='*** 2'),
    partial(is_upflow,geo=geo,convention='*** 1'),
    partial(is_upflow_rech,geo=geo,convention='*** 3'),
    partial(is_side_rech,geo=geo,convention='*****')
    ]

    print('+++++', nsdat.filename)
    count = check_geners(nsdat.generatorlist, g_types)
    for t in sorted(count.keys()):
        print('%20s = %i' % (t,count[t]))
    print('Total # of generators is ', len(nsdat.generatorlist))

    print('+++++', prdat.filename)
    count = check_geners(prdat.generatorlist, g_types)
    for t in sorted(count.keys()):
        print('%20s = %i' % (t,count[t]))
    print('Total # of generators is ', len(prdat.generatorlist))



import re
import numpy as np

def get_block_history(bname, lsts, cols=None):
    """ return history tables keyed by block name
    bname can be either a list, or a regular expression string
    """
    if cols is None:
        cols = ['Pressure']
    if not isinstance(lsts,list):
        lsts = [lsts]
    if isinstance(bname, list):
        # change list og block names into regular expression
        bname = '(' + '|'.join(bname) + ')'

    full_table = {}
    full_bcns = []
    for lst in lsts:
        print('in listing file %s:' % lst.filename)
        # create zeros for each matching (b,g,cn) in current listing
        bis, bs = [], [] # indices in gener and element tables
        blk_hist, blk_time, collated_bs = {}, {}, [] # collate potential multiple (b,g)s into a single one
        for i,b in enumerate(lst.element.row_name):
            if re.match(re.compile(bname), b):
                bis.append(i)
                bs.append(b)
                for cn in cols:
                    if (b,cn) not in blk_hist:
                        collated_bs.append((b,cn))
                        blk_hist[(b,cn)] = np.zeros_like(lst.fulltimes)
                        blk_time[(b,cn)] = np.copy(lst.fulltimes)
        # sum through tables in current listing
        for j,t in enumerate(lst.fulltimes):
            lst.index = j
            for cn in cols:
                for i,b in zip(bis, bs):
                    blk_hist[(b,cn)][j] = blk_hist[(b,cn)][j] + lst.element[cn][i]
        for k in collated_bs:
            if k not in full_table:
                print('    (new) found matching block %s' % str(k))
                full_bcns.append(k)
                full_table[k] = (blk_time[k], blk_hist[k])
            else:
                print('          found matching block %s' % str(k))
                full_table[k] = (np.concatenate((full_table[k][0], blk_time[k])),
                                 np.concatenate((full_table[k][1], blk_hist[k])))
    return full_table, full_bcns


def get_gener_history(bname, gname, lsts, cols=None, silent=True):
    """ return history tables keyed by (block name, gener name, column name)
    bname and gname are strings that support regular expression.

    If bname and gname are simple strings, they are used as as regular
    expression to match any geners that match them.  A tuple of (full_table,
    full_bgcns) is returned.  full_bgcns is a list of (blockname, genername,
    columnname).  Full table is a dict using the (b, g, cn) as key.  Each item
    is a tuple of times and values.

    If bname and gname are supplied as two lists (needs to match length),
    full_bgcns will be a list of tuple that is identical to zip(bname, gname).
    Any of these entries will have zero values filled with the same length as
    lst.fulltimes.

    Values of repeated (b,g) generators are summed together.  Enthalpy is
    averaged by generation rate.  If the gener has flowrate of zero, it will use
    the enthalpy value from the listing table.  Usually AUT2 (DMAK/DMAT etc)
    calculate the enthalpy value anyway even if not flowing at the time.

    Internally it always sums up Generation rate and averages Enthalpy, even if
    they are absent from user specified cols.  This greatly simplifies the
    unique averaging enthalpy, which is often required anyway.
    """
    if not isinstance(lsts,list):
        lsts = [lsts]

    if isinstance(bname, list) and isinstance(gname, list) and len(bname) == len(gname):
        # pad zeros even if (b,g) not exist in current listing
        add_bgs = list(zip(bname,gname))
    else:
        # if bname and gname are strings, assume regular expression
        add_bgs = []
        rebn, regn = re.compile(bname), re.compile(gname)

    rate_cname, stm_cname, enth_cname = None, None, None
    for cn in lsts[0].generation.column_name:
        if cn.lower().startswith('generation ra'):
            rate_cname = cn
        if cn.lower().startswith('steam sep'):
            stm_cname = cn
        if cn.lower().startswith('enthal'):
            enth_cname = cn
    if cols is None:
        cols = [rate_cname]

    full_table = {}
    full_bgcns = []
    for lst in lsts:
        if not silent:
            print('in listing file %s:' % lst.filename)
        # create zeros for each matching (b,g,cn) in current listing
        gen_hist, gen_time = {}, {} # used interally
        collated_bgcns = [] # collate potential multiple (b,g)s into a single one

        # have value zeros even if specified (b,g) not in current t2listing
        for b,g in add_bgs:
            for cn in cols:
                if (b,g,cn) not in gen_hist:
                    collated_bgcns.append((b,g,cn))
                    gen_hist[(b,g,cn)] = np.zeros_like(lst.fulltimes)
                    gen_time[(b,g,cn)] = np.copy(lst.fulltimes)
            if (b,g,rate_cname) not in gen_hist:
                gen_hist[(b,g,rate_cname)] = np.zeros_like(lst.fulltimes)
            if (b,g,enth_cname) not in gen_hist:
                gen_hist[(b,g,enth_cname)] = np.zeros_like(lst.fulltimes)
        # actually exists in current t2listing
        gis, bgs = [], [] # indices in gener table
        for i,(b,g) in enumerate(lst.generation.row_name):
            found = False
            if add_bgs:
                # if one of specified list
                if (b,g) in add_bgs:
                    found = True
            else:
                if re.match(rebn, b) and re.match(regn, g):
                    found = True
            if found:
                gis.append(i)
                bgs.append((b,g))
                for cn in cols:
                    if (b,g,cn) not in gen_hist:
                        collated_bgcns.append((b,g,cn))
                        gen_hist[(b,g,cn)] = np.zeros_like(lst.fulltimes)
                        gen_time[(b,g,cn)] = np.copy(lst.fulltimes)
                if (b,g,rate_cname) not in gen_hist:
                    gen_hist[(b,g,rate_cname)] = np.zeros_like(lst.fulltimes)
                if (b,g,enth_cname) not in gen_hist:
                    gen_hist[(b,g,enth_cname)] = np.zeros_like(lst.fulltimes)
        # sum/average through tables in current listing
        for j,t in enumerate(lst.fulltimes):
            lst.index = j
            for i,(b,g) in zip(gis, bgs):
                # do massflow and average enthalpy separately
                total_heat = gen_hist[(b,g,rate_cname)][j] * gen_hist[(b,g,enth_cname)][j]
                total_heat += lst.generation[rate_cname][i] * lst.generation[enth_cname][i]
                total_mass = gen_hist[(b,g,rate_cname)][j] + lst.generation[rate_cname][i]
                if abs(total_mass) < 1.0e-10:
                    # avg_enth = 0.0
                    avg_enth = lst.generation[enth_cname][i]
                else:
                    avg_enth = total_heat / total_mass
                gen_hist[(b,g,enth_cname)][j] = avg_enth
                gen_hist[(b,g,rate_cname)][j] = total_mass
                for cn in cols:
                    if cn in [rate_cname, enth_cname]:
                        # already done separately above
                        pass
                    else:
                        # simply add everything else
                        gen_hist[(b,g,cn)][j] = gen_hist[(b,g,cn)][j] + lst.generation[cn][i]
        for k in collated_bgcns:
            # k is (block name, gener name, column name)
            if k not in full_table:
                if not silent:
                    print('    (new) found matching gener %s' % str(k))
                full_bgcns.append(k)
                full_table[k] = (gen_time[k], gen_hist[k])
            else:
                if not silent:
                    print('          found matching gener %s' % str(k))
                full_table[k] = (np.concatenate((full_table[k][0], gen_time[k])),
                                 np.concatenate((full_table[k][1], gen_hist[k])))
    return full_table, full_bgcns

def get_gener_history_v0(bname, gname, lsts, cols=None):
    """ return history tables keyed by (block name, gener name, column name)
    bname and gname are strings that support regular expression

    Values of repeated (b,g) generators are summed together.  Hence column
    name such as Enthalpy won't work properly
    """
    if cols is None:
        cols = ['Generation rate']
    if not isinstance(lsts,list):
        lsts = [lsts]

    full_table = {}
    full_bgcns = []
    for lst in lsts:
        print('in listing file %s:' % lst.filename)
        # create zeros for each matching (b,g,cn) in current listing
        gis, bgs = [], [] # indices in gener and element tables
        gen_hist, gen_time, collated_bgcns = {}, {}, [] # collate potential multiple (b,g)s into a single one
        for i,(b,g) in enumerate(lst.generation.row_name):
            if re.match(re.compile(bname), b) and re.match(re.compile(gname), g):
                gis.append(i)
                bgs.append((b,g))
                for cn in cols:
                    if (b,g,cn) not in gen_hist:
                        collated_bgcns.append((b,g,cn))
                        gen_hist[(b,g,cn)] = np.zeros_like(lst.fulltimes)
                        gen_time[(b,g,cn)] = np.copy(lst.fulltimes)
        # sum through tables in current listing
        for j,t in enumerate(lst.fulltimes):
            lst.index = j
            for cn in cols:
                for i,(b,g) in zip(gis, bgs):
                    gen_hist[(b,g,cn)][j] = gen_hist[(b,g,cn)][j] + lst.generation[cn][i]
        for k in collated_bgcns:
            if k not in full_table:
                print('    (new) found matching gener %s' % str(k))
                full_bgcns.append(k)
                full_table[k] = (gen_time[k], gen_hist[k])
            else:
                print('          found matching gener %s' % str(k))
                full_table[k] = (np.concatenate((full_table[k][0], gen_time[k])),
                                 np.concatenate((full_table[k][1], gen_hist[k])))
    return full_table, full_bgcns



""" CONFIG 
"""


class Singleton(object):
    __single = None # the one, true Singleton

    def __new__(classtype, *args, **kwargs):
        # Check to see if a __single exists already for this class
        # Compare class types instead of just looking for None so
        # that subclasses will create their own __single objects
        if classtype != type(classtype.__single):
            classtype.__single = object.__new__(classtype, *args, **kwargs)
        return classtype.__single

        
#class config(Singleton):
class config(object):
    def __init__(self,filename=''):
        self.empty()
        if filename: self.read_from_file(filename)
    def empty(self):
        self._config_entries = {}
        
    def read_from_file(self,filename=None):
        """ read config from a whole file, until [END], if filename
            is not given, user will be prompted for a filename."""
        if filename==None:
            filename=raw_input(' Configuration file (.cfg) name? ')
        print(" Reading file", filename.strip(), "for configurations...")
        
        cfgfile=open(filename)
        self._read(cfgfile)    
        cfgfile.close()
        
    def read_from_file_section(self,file_with_cfg):
        """ only read config entries from the current location, 
            file_with_cfg should be already opened, and reading will
            be terminated once [END] is reached, the calling program 
            can continue read the rest of the file. """
        self._read(file_with_cfg)
        
    def _read(self,cfgfile):
        finished=False
        while not finished:
            line = cfgfile.readline()
            if line:
                if line.strip()=='': continue
                if line.strip()[0]=='!': continue
                if line.strip()[0]=='#': continue
                if line.strip()[0]=='[':
                    ikeyend=line.find(']')
                    keyword=line[1:ikeyend]
                    if keyword=='END': break
                    self._config_entries[keyword]=[]
                else: 
                    self._config_entries[keyword].append(line.rstrip('\n\r'))
            else: finished = True
    def add_value(self,keyword,value):
        """ add value manually, value can be a single value, or a list, 
            or whatever object, as long as you know what it is when you
            get it out, if keyword exist, the value wil be appended """
        if keyword in self._config_entries:
            try:
                self._config_entries[keyword].append(value)
            except:
                self._config_entries[keyword] = [self._config_entries[keyword], value]
        else:
            self._config_entries[keyword] = [value]
        
    def get_value(self,keyword):
        if len(self._config_entries[keyword]) == 0:
            return ''
        else:
            return self._config_entries[keyword][0]
        
    def get_list(self,keyword):
        return self._config_entries[keyword]

    def check_optional(self,keyword):
        return keyword in self._config_entries.keys()




""" Module for generating report tables of scenario specification.

This module replies on .json files generated along side scenario simulations.
Tables in LaTeX format is genreated for use in the figure report.
"""
import json
from copy import deepcopy

from gimfu.figure_report import escape_latex_special_chars

class CleanTable(object):
    """ Generating clean table where repeated cell is empty if content is the
        same as previous row.

    Attributes:
        headers (list): List of headers strings.
        table (list): List of table rows.
    """
    def __init__(self, headers=None, omit_repeats=[], widths=None):
        """
        Args:
            headers (list): List of headers strings.

            omit_repeats (list): List of columns (headers) that should omit
            content if cell is the same as previous row.
        """
        if widths:
            if len(widths) != len(headers):
                raise ValueError("Widths must have the same length as headers.")
        self.headers = headers
        self.widths = widths
        self.omit_repeats = [self.headers.index(h) for h in omit_repeats]
        self.table = []
        # the following are automatically generated/formatted/refreshed
        self.detected_widths = [len(h) for h in self.headers]
        self._clean_headers = self.headers
        self._clean_table = None

    def add_row(self, row):
        """ Add a row to the table.

        Args:
            row (list): List of row strings.
        """
        self.table.append(row)
        self._clean_table = None # reset clean table

    @property
    def clean_table(self):
        # update clean table
        if self._clean_table is None:
            self.detected_widths = [len(h) for h in self.headers]
            self._clean_table = []
            prev_row = None
            for row in self.table:
                row_copy = deepcopy(row)
                if prev_row:
                    for i, cell in enumerate(row_copy):
                        self.detected_widths[i] = max(self.detected_widths[i], len(cell))
                        if cell == prev_row[i] and (i in self.omit_repeats):
                            # row_copy[i] = ' ' * len(prev_row[i])
                            row_copy[i] = ''
                self._clean_table.append(row_copy)
                prev_row = row
        # fix and align column widths
        if self.widths:
            final_widths = self.widths
        else:
            final_widths = self.detected_widths
        self._clean_headers = [h.ljust(w) for h, w in zip(self.headers, final_widths)]
        # self._clean_headers = [h.strip().ljust(w) for h, w in zip(self.headers, final_widths)]
        for row in self._clean_table:
            for i, cell in enumerate(row):
                row[i] = cell.ljust(final_widths[i])
                # row[i] = cell.strip().ljust(final_widths[i])
        return self._clean_table

    @property
    def clean_headers(self):
        # ensure clean header is updated
        clean_table = self.clean_table
        return self._clean_headers

    def dumps_latex(self, print_header=True, hline=True, caption=''):
        """ return lines in LaTeX format. """
        from tabulate import tabulate
        table = tabulate(self.clean_table, headers=self.clean_headers,
                         tablefmt='latex_longtable',
                         disable_numparse=True)
        if caption:
            caption_code = "\\caption{%s}\n" % escape_latex_special_chars(caption)
            # caption_code += "\n".join([
            #     '\\ttfamily',
            #     '\\tiny',
            # ])
            # only replace the first occurrence
            table = table.replace("\\end{longtable}", caption_code + "\n\\end{longtable}", 1)

        lines = [
            '',
            # '\\begin{table}',
            '\\ttfamily',
            '\\tiny',
            # '\\centering',
            table,
            # '\\caption{%s}' % escape_latex_special_chars(caption),
            # '\\end{table}',
            '',
            ]
        return '\n'.join(lines)

    def dumps_txt(self, print_header=True, hline=True, vsep=' | '):
        """ return lines pure text or on screen. """
        table = self.clean_table
        headers = self.clean_headers
        hline_width = sum([len(h) for h in headers]) + (len(self.headers)-1) * len(vsep)
        lines = []
        if headers and print_header:
            lines.append(vsep.join(headers))
        if hline:
            lines.append('-' * hline_width)
        for row in table:
            lines.append(vsep.join(row))
        if hline:
            lines.append('-' * hline_width)
        return '\n'.join(lines)


def collate_well_stacks(spec_json):
    """ Collate well stacks from all simulations in a scenario.

    Returns:
        well_stack_specs: dict of well stack specifications. Primarily used for
            generating a table with stack_spec_table().

        well_geners: dict of well generators. eg. {
            'WK1': [('abc 1', 'WW  1'), ('abc 2', 'WW  1')],
            'WK2': [('xyz 5', 'TT  2'), ('xyz 5', 'WW  2')],
        }

    Args:
        spec_json: json-like object of scenario specification.
    """
    well_stacks = []
    well_stack_specs = {}
    well_geners = {}
    simulations = spec_json["simulations"]
    for sim in simulations:
        for spec in sim['specs']:
            if spec:
                for stack in spec['well_stacks']:
                    if stack:
                        well_stacks.append(stack)

                        # well_geners
                        stack_spec = spec['specifications'][stack]
                        for well in stack_spec['wells']:
                            if well not in well_geners:
                                well_geners[well] = []
                            for gener in stack_spec['geners'][well]:
                                if gener not in well_geners[well]:
                                    well_geners[well].append(gener)
                            # well_geners[well] += stack_spec['geners'][well]

                        # well_stack_specs
                        well_stack_specs.update(spec['specifications'])

    return well_stack_specs, well_geners


def stack_schedule_table(spec_json):
    """ Generate a table for well stacks used in a scenario.

    Args:
        spec_json: json-like object of scenario specification.
    """
    scenario = spec_json
    simulations = spec_json["simulations"]
    table = CleanTable(headers=['Scenario', 'Dates', 'Well Stack'],
                       omit_repeats=['Scenario', 'Dates'],
                       # widths=[10, 20, 50],
                       )
    str_sname = "%5s" % scenario['name']
    for sim in simulations:
        str_dates = "%6.1f - %6.1f" % (sim['tstart_yr'], sim['tstop_yr'])
        well_stacks = []
        for spec in sim['specs']:
            if spec:
                for stack in spec['well_stacks']:
                    if stack:
                        well_stacks.append(stack)
        for stack in well_stacks:
            table.add_row([str_sname, str_dates, stack])
    return table


def stack_spec_table(stack_spec_json, abbrev={}):
    """ Generate a table for a well stack specification.

    Args:
        stack_spec_json: json-like object of well stack specification.
        abbrev: dict of abbreviations to replace long names.
    """
    table = CleanTable(headers=[
                           'Well',
                           'Geners',
                           'Blocks',
                           'Clean t/hr',
                           'Cap t/hr',
                           'Ecut kJ/kg',
                           'Psep bar',
                           'Cap by',
                           'Deliv.~Curve',
                           ],
                       # omit_repeats=[],
                       # widths=[50, 10, 50],
                       )
    def fmt(fmt, a):
        if a is not None:
            if a in abbrev:
                return fmt % abbrev[a]
            return fmt % a
        else:
            return ''
    for well in stack_spec_json['wells']:
        all_bnames = sorted(set([g[0] for g in stack_spec_json['geners'][well]]))
        blocks_str = ','.join(all_bnames)
        all_gnames = sorted(set([g[1] for g in stack_spec_json['geners'][well]]))
        geners_str = ','.join(all_gnames)
        table.add_row([
            fmt('%s'   , well),
            fmt('%s', geners_str),
            fmt('%s', blocks_str),
            fmt('%5.1f', stack_spec_json['clean_rate_t/hr'][well]),
            fmt('%5.1f', stack_spec_json['cap_t/hr'][well]),
            fmt('%6.0f', stack_spec_json['e_cutoff_kJ/kg'][well]),
            fmt('%5.2f', stack_spec_json['p_sep_bara'][well]),
            fmt('%s'   , stack_spec_json['cap_spec_source'][well]),
            fmt('%s'   , stack_spec_json['deliv_curve'][well]),
            ])
    return table

def well_geners_table(well_geners, custom_sort_key=None):
    """ Generate a table for well generators.

    Args:
        well_geners: dict of well generators.
    """
    table = CleanTable(headers=['Well', 'Generator', ], omit_repeats=['Well'], )
    for well in sorted(well_geners.keys(), key=custom_sort_key):
        for gener in well_geners[well]:
            table.add_row([well, '(%5s:%5s)' % tuple(gener), ])
    return table

def well_blocks_geners_table(well_geners, custom_sort_key=None):
    """ Generate a table for well blocks.

    Args:
        well_geners: dict of well generators.
    """
    table = CleanTable(headers=['Well', 'Model Blocks', 'Geners'], )
    for well in sorted(well_geners.keys(), key=custom_sort_key):
        all_bnames = sorted(set([g[0] for g in well_geners[well]]))
        all_gnames = sorted(set([g[1] for g in well_geners[well]]))
        table.add_row([well, ', '.join(all_bnames), ', '.join(all_gnames), ])
    return table

def well_blocks_table(well_geners, custom_sort_key=None):
    """ Generate a table for well blocks.

    Args:
        well_geners: dict of well generators.
    """
    table = CleanTable(headers=['Well', 'Model Blocks'], )
    for well in sorted(well_geners.keys(), key=custom_sort_key):
        all_bnames = sorted(set([g[0] for g in well_geners[well]]))
        table.add_row([well, ', '.join(all_bnames), ])
    return table


if __name__ == "__main__":
    with open('scenario_spec.json', 'r') as f:
        spec_json = json.load(f)

    # abbreviate long names
    abbrev = {
        'steam-cutoff-2.00-bar': 'Steam (2.0 bar WHP)',
        './make_pcutoff_curves_TH12_50WHP/make_Pcutoff_curve.out': 'TH12  (5.0 bar WHP)',
    }

    # Custom sort key function
    def custom_sort_key(s):
        # Return a tuple where the first element determines if 'WK' is present or not
        # and the second element is the string itself for normal sorting.
        return (0 if s.startswith("WK") else 1, s)

    tables = []
    # tables.append(stack_schedule_table(spec_json))

    well_stack_specs, well_geners = collate_well_stacks(spec_json)
    tables.append(stack_spec_table(well_stack_specs['THIPOI_PH2x_137sep'],
                                   abbrev=abbrev))
    # tables.append(well_geners_table(well_geners, custom_sort_key=custom_sort_key))
    tables.append(well_blocks_geners_table(well_geners, custom_sort_key=custom_sort_key))
    # tables.append(well_blocks_table(well_geners, custom_sort_key=custom_sort_key))

    for table in tables:
        print('\n\n')
        print(table.dumps_txt())
        # print(table.dumps_latex())





""" provides interface to easily change the content of a existing GNUPlot
 .gpl file.  Not full gpl feature supported, most of the properties is
 processed like normal string.  See __main__ for example on how to use """

from pathlib import Path

def unquote(text):
    t = text.strip()
    if len(t) < 2: raise Exception("Error at unquoting text: %s" % text)
    if t[0] not in '"'+"'": raise Exception("Error at unquoting text: %s" % text)
    if t[0] != t[-1]: raise Exception("Error at unquoting text: %s" % text)
    return t[1:-1]
def quote(text):
    q = "'"
    if q in text: q = '"'
    return q+text+q

def cleanup_multi_lines(multi_line_string):
    """ clean up the line continuation of \ in gpl files, the input
    should be the original stream of lines, with line break kept, so
    line continuation can be detected """
    lines = multi_line_string.splitlines()
    cleaned = ''
    for line in lines [:-1]:
        cleaned += line[:-1]
    return cleaned + lines[-1]

def string_eat_group(astring):
    """ will extract the string inside the grouping marks, returning
    two strings, first being the content inside the marks, second being
    the left over string.  NOTE the first char of the original string
    must be the opening of grouping character. """
    groupchar = {'"':'"',"'":"'",'{':'}','[':']','<':'>','(':')','`':'`','$':'$'}
    c = astring[0]
    iend = astring[1:].find(groupchar[c])
    if iend == -1: raise Exception('matching grouping char of %s is not found in: %s' % (c,astring))
    return (astring[1:iend+1],astring[iend+2:])

def string_eat_until(astring,until=' '):
    """ extract part of the string up to specified word or any of the
    words if a list is given, and return the leftover string as well. """
    astring = astring.lstrip()
    if type(until) == type([]):
        iend = [astring.find(u) for u in until if astring.find(u) != -1]
        if len(iend) == 0: return(astring,'')
        return (astring[:min(iend)],astring[min(iend):])
    else:
        iend = astring.find(until)
        if iend == -1: return(astring,'')
        return (astring[:iend],astring[iend:])

class series(object):
    def __init__(self,cmdline):
        # these two will always be identical
        self.datafile, self.function = '', ''
        self._actual_cmd = ''
        # only has value if specified or read in-place from .gpl
        self.data = []
        self.axes = None
        self.title = None
        self.withstyle = None
        line = cmdline.strip()
        (command,line) = string_eat_until(line,[' axes',' title',' notitle',' with'])
        self.datafile, self.function = command, command
        # parse the rest
        while line.strip() != '':
            line = line.strip()
            if line.find(' ') == -1:
                # single word left
                kwrd = line
                value = ''
            else:
                (kwrd,line) = string_eat_until(line.strip(),' ')
                (value,line) = string_eat_until(line.strip(),[' axes',' title',' notitle',' with'])
                # print '@@@@@'+ kwrd+'@'+value+'@'
            if kwrd == 'axes':       self.axes = value
            elif kwrd == 'title':    self.title = unquote(value)
            elif kwrd == 'notitle':  self.title = ''
            elif kwrd == 'with':    self.withstyle = value
        # fix bplot.exe bug by overwrite this early on
        self._fix_bplot_no_ls_issue()

    def __setattr__(self,name,value):
        """ update self._actual_cmd according to datafile OR function """
        if name == 'datafile':
            if value is None:
                self._actual_cmd = self.function or ''
            else:
                if value == '': value = "'-'"
                # add qutation for user if forget
                if value[0] not in '"'+"'": value = "'" + value + "'"
                self._actual_cmd = value
        if name == 'function':
            if value is None:
                self._actual_cmd = self.datafile or ''
            else:
                self._actual_cmd = value
        # proceed as normal, for all
        self.__dict__[name] = value

    def get_with(self):
        if self.withstyle is None:
            return None
        else:
            # always the first option, not optional in gnuplot I think
            return self.withstyle.split()[0]
    def set_with(self, style):
        """
        style is a string, must be one of (see gnuplot manual):
            lines dots steps errorbars xerrorbar xyerrorlines
            points impulses fsteps errorlines xerrorlines yerrorbars
            linespoints labels histeps financebars xyerrorbars yerrorlines
            surface vectors parallelaxes
        or
            boxes boxplot ellipses image
            boxerrorbars candlesticks filledcurves rgbimage
            boxxyerrorbars circles histograms rgbalpha pm3d
        or
            table
        """
        if style is None:
            self.withstyle = style
            return
        if self.withstyle is None:
            self.withstyle = style
        sps = self.withstyle.split()
        self.withstyle = ' '.join([style] + sps[1:])

    def get_ls(self):
        """ get 'ls' value from the 'with' command """
        sps = self.withstyle.split()
        i = None
        for k in ['ls', 'linestyle']:
            try:
                i = sps.index(k)
            except ValueError:
                continue
        if i is None:
            return None
        else:
            return sps[i+1]
    def set_ls(self, ls):
        """ set 'ls' value from the 'with' command """
        if self.withstyle is None:
            self.set_style('linespoints')
        sps = self.withstyle.split()
        i = None
        for k in ['ls', 'linestyle']:
            try:
                i = sps.index(k)
            except ValueError:
                continue
        if i is None:
            if ls is None:
                return
            else:
                # no ls specified yet, insert at the start
                self.withstyle = ' '.join(sps[:1] + ['ls', str(ls)] + sps[2:])
        else:
            if ls is None:
                self.withstyle = ' '.join(sps[:i] + sps[i+2:])
            else:
                self.withstyle = ' '.join(sps[:i+1] + [str(ls)] + sps[i+2:])

    def _fix_bplot_no_ls_issue(self):
        """ fix AY bplot.exe's bug of missing linestyle/ls in the with command.
        gnuplot's 'with' command:
            with <style> { {linestyle | ls <line_style>}
                           | {{linetype | lt <line_type>}
                              {linewidth | lw <line_width>}
                              {linecolor | lc <colorspec>}
                              {pointtype | pt <point_type>}
                              {pointsize | ps <point_size>}
                              {fill | fs <fillstyle>}
                              {nohidden3d} {nocontours} {nosurface}
                              {palette}}
                         }

        """
        if self.withstyle is not None:
            sps = self.withstyle.split()
            # the second item in the command SHOULD NOT be a number
            # fix this bplot.exe bug by inserting 'ls'
            if len(sps) > 1 and sps[1].isdigit():
                sps.insert(1, 'ls')
            self.withstyle = ' '.join(sps)

    def write(self):
        command = ''
        command += self._actual_cmd
        if self.axes is not None:
            command += ' axes '+ self.axes
        if self.title is not None:
            command += ' title '+ quote(self.title)
        if self.withstyle is not None:
            # seems gnuplot after 5.0 no longer accepts 'linespoints 1', but need 'linespoints ls 1'
            if self.withstyle.strip().startswith('linespoints'):
                if 'ls' not in self.withstyle:
                    self.withstyle = self.withstyle.replace('linespoints lt', 'linespoints ls')
            command += ' with '+ self.withstyle
        return command
    command_string = property(write)

    def __repr__(self):
        return '<gpl.series object>: %s' % self.command_string

    def read_series_data(self,openedfile):
        if self.datafile[0] in '"'+"'":
            fn,tmp = string_eat_group(self.datafile.strip())
            if fn == '-':
                # read until 'e'
                while True:
                    line = openedfile.readline()
                    if line.strip() == 'e': break
                    self.data.append([float(x) for x in line.split()])

    def write_series_data(self,openedfile):
        if self.datafile[:3] in ["'-'", '"-"']:
            # from os import linesep
            linesep = '\n'
            for xs in self.data:
                openedfile.write(' '.join([('%20.13e' % x) for x in xs]) + linesep)
            # for a,b in self.data:
            #     openedfile.write('%20.13e %20.13e%s' % (a,b,linesep))
            openedfile.write('e'+linesep)

    def reverse_yvalue(self):
        for d in self.data:
            d[1] = -d[1]


class gplfile(object):
    option_is_text = {
    'terminal'   : False ,
    'title'      : True  ,
    'output'     : True  ,
    'key'        : False ,
    'xlabel'     : True  ,
    'ylabel'     : True  ,
    'logscale'   : False ,
    'xrange'     : False ,
    'yrange'     : False ,
    'xtics'      : False ,
    'ytics'      : False ,
    'mxtics'     : False ,
    'mytics'     : False ,
    'grid'       : False
     }
    def __init__(self,filename):
        self.series=[]
        self.read(filename)
        self.filename = filename

    def read(self,filename):
        self._cmdlines = []
        f = open(filename,'r')
        cmdline = ''
        line = ' '
        while line:
            line = f.readline()
            # record empty line, to avoid next detection fail
            if line.strip() == '':
                self._cmdlines.append(line)
                continue
            # combine multi-lines into one cmdline if ends with \
            cmdline = line
            while line.rstrip('\r\n')[-1] == '\\':
                line = f.readline()
                cmdline += line
            # processing a cmdline
            words = cleanup_multi_lines(cmdline).split()
            if words[0] == 'set':
                if words[1] in gplfile.option_is_text.keys():
                    istart = cmdline.find(words[1]) + len(words[1]) + 1
                    self.__dict__[words[1]] = cmdline[istart:].rstrip().strip('"'+"'")
            elif words[0] == 'unset':
                if words[1] in gplfile.option_is_text.keys():
                    if words[1] in self.__dict__: self.__dict__[words[0]] = None
            elif words[0] == 'plot':
                cmdline = cleanup_multi_lines(cmdline.lstrip())
                self.series = [series(a) for a in cmdline[4:].split(',')]
                # special to plot command, needs to consume the expected data
                for s in self.series:
                    s.read_series_data(f)
                # ignore all entries after plot (they won't affect plot anyway
                break
            # always record cmdline
            self._cmdlines.append(cmdline)
        f.close()

    def write(self,filename = None):
        if filename is None:
            filename = self.filename
        linesep = '\n'
        f = open(filename,'w')
        # print those other options, before plot command
        for orig_line in self._cmdlines:
            words = cleanup_multi_lines(orig_line).split()
            to_write = orig_line
            if len(words) > 0:
                if words[0] in ['set','unset']:
                    if words[1] in gplfile.option_is_text.keys():
                        if words[1] in self.__dict__:
                            if self.__dict__[words[1]] is None:
                                to_write = 'unset '+words[1]+linesep
                            else:
                                to_write = 'set '+words[1]+' '+self._auto_quote_value(words[1])+linesep
            f.write(to_write)
        ### print plot command
        if len(self.series) < 1: raise Exception('<gpl.gplfile object> write() failed, there must be at least one series to plot')
        ### plot command
        f.write('plot \\'+linesep)
        for s in self.series[:-1]:
            f.write(s.command_string+',\\'+linesep)
        f.write(self.series[-1].command_string+linesep)
        ### write data section
        for s in self.series:
            s.write_series_data(f)
        ### end comment
        f.write(linesep)
        from datetime import datetime
        f.write('# file writted by gpl.py '+datetime.now().__str__()+linesep)
        f.write(linesep)
        f.close()

    def set(self,option,value):
        if option in gplfile.option_is_text.keys():
            if option not in self.__dict__:
                self._cmdlines.append("set %s %s" % (option,value))
            self.__dict__[option] = value
    def unset(self,option):
        if option in gplfile.option_is_text.keys():
            self.__dict__[option] = None

    def _auto_quote_value(self,option):
        if option not in self.__dict__.keys(): return None
        value = str(self.__dict__[option])
        if value is None: return ''
        if value == '': return ''
        if gplfile.option_is_text[option]:
            if '"' in value: q = "'"
            else: q = '"'
            if option == 'output': q = "'"
            value = q+value+q
        return value

    def __repr__(self):
        allopts = []
        for opt in sorted(gplfile.option_is_text.keys()):
            if opt not in self.__dict__.keys(): continue
            val = self._auto_quote_value(opt)
            allopts.append('set %s %s' % (opt,val))
        return '\n'.join([
            '<gpl.gplfile object> with overwriting options:'
            ] + allopts)

    def remove_abs_path(self):
        self.output = str(Path('.') / Path(self.output).name)
        # i = self.output.rfind('\\')
        # self.output = '.'+self.output[i:]

    def reverse_yvalue(self):
        for s in self.series:
            s.reverse_yvalue()

    def auto_xrange_min(self, r):
        r_half = r / 2.0
        sss = []
        for s in self.series:
            xs,ys = zip(*(s.data))
            sss = sss + list(xs)
        smin, smax = min(sss), max(sss)
        if (smax - smin) < r:
            smean = (smin + smax) / 2.0
            self.set('xrange', '[*<%f:%f<*]' % (smean - r_half, smean + r_half))

    def auto_yrange_min(self, r):
        r_half = r / 2.0
        sss = []
        for s in self.series:
            xs,ys = zip(*(s.data))
            sss = sss + list(ys)
        smin, smax = min(sss), max(sss)
        if (smax - smin) < r:
            smean = (smin + smax) / 2.0
            self.set('yrange', '[*<%f:%f<*]' % (smean - r_half, smean + r_half))

    def set_filetype(self, filetype):
        """ convnience function of changing between .eps and .png
        """
        terms = {
            'png': 'pngcairo enhanced font "Arial,18" size 1000,750',
            'eps': 'postscript landscape enhanced colour "Arial" 18',
        }
        self.terminal = terms[filetype]
        for ext in terms.keys():
            self.output = self.output.replace('.'+ext, '.'+filetype)

    def set_series_with(self, indices, linespoints):
        """ overwrite series to show either as points only, line only or both

        indices: is a list of indices, if an indix is out of current plot's
                 range, it will be ignored
        linespoints: should be 'lines', 'points' or most commonly 'linespoints'
        """
        if linespoints not in['lines', 'points', 'linespoints']:
            raise Exception
        for i,s in enumerate(self.series):
            if i in indices:
                if self.withstyle.strip().startswith('linespoints'):
                    if 'ls' not in self.withstyle:
                        self.withstyle = self.withstyle.replace('linespoints', 'linespoints ls')


    def set_custom_linestyles(self, linewidth=2, pointsize=1.2):
        # be careful to add \n to end each line
        additional_gpl_cmds = [
            " \n",
            "set style line  1 lc rgb '#FF0000' lt 1 lw %f pt  1 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line  3 lc rgb '#0000FF' lt 1 lw %f pt  6 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line  4 lc rgb '#008000' lt 1 lw %f pt  8 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line  8 lc rgb '#FF8C00' lt 1 lw %f pt 10 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line 12 lc rgb '#00BFFF' lt 1 lw %f pt  2 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line 13 lc rgb '#696969' lt 1 lw %f pt 12 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line 14 lc rgb '#FF00FF' lt 1 lw %f pt 14 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line 16 lc rgb '#0000FF' lt 1 lw %f pt  1 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line 27 lc rgb '#008000' lt 1 lw %f pt  6 ps %f # pi -1 \n" % (linewidth, pointsize),
            "set style line 36 lc rgb '#FF8C00' lt 1 lw %f pt  8 ps %f # pi -1 \n" % (linewidth, pointsize),
            " \n",
        ]
        for cmd in additional_gpl_cmds:
            self._cmdlines.append(cmd)
        # NOTE:
        # a series with bplot.exe generated 'with linespoints 1' will be
        # automatically updated to 'with linespoints ls 1', so GNUPlot 5.0+ can
        # deal with it correctly


if __name__ == "__main__":
    g = gplfile('test.gpl')
    s = g.series[0]


    # g.title = g.title + 'XXX'
    # g.series[1].title = g.series[1].title + 'XXX'

    # print g.output
    # g.set('yrange','[0:100]')
    # g.unset('xrange')
    # g.remove_abs_path()
    # print g.output

    # g.write('test1.gpl')







import os
import shutil
import subprocess
import gimfu.gpl
from copy import deepcopy

import pint
Unit = pint.UnitRegistry()
Unit.Q = Unit.Quantity

titlepage = (
    "\\begin{titlepage}\n"
    "    \\begin{center}\n"
    "\n"
    "    \\begin{minipage}[c]{1.0\\textwidth}\n"
    "        \\raisebox{-0.5\\height}{\\includegraphics[width=0.50\\textwidth]{../UniServices.eps}}\n"
    "        \\hfill\n"
    "        \\raisebox{-0.5\\height}{\\includegraphics[width=0.30\\textwidth]{../The_University_of_Auckland.eps}}\n"
    "    \\end{minipage}\n"
    "\n"
    "    \\vspace*{-1cm}\n"
    "\n"
    "    \\Huge\n"
    "    \\textsc{\n"
    "        Wair\={a}kei-Tauhara Modelling Report\n"
    "    }\n"
    "\n"
    "    \\LARGE\n"
    "    \\textsc{%s}\n"
    "\n"
    "    \\vspace*{1.5cm}\n"
    "\n"
    "    \\huge\n"
    "    \\textsc{\n"
    "    by \\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    Angus Yeh,\\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    Prof. Michael O'Sullivan,\\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    Dr John O'Sullivan,\\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    Ethan Chabora, Paul Abrasaldo\\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    and Warren Mannington\\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    (Contact Energy Ltd.)\n"
    "    }\n"
    "\n"
    "    \\vspace*{1.0cm}\n"
    # "    \\vfill\n"
    "\n"
    "    \\huge\n"
    "    \\textsc{\n"
    "    Uniservices and \\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    Department of Engineering Science \\\\\n"
    "    \\vspace*{0.1cm}\n"
    "    University of Auckland\n"
    "    }\n"
    "\n"
    "    \\vspace*{0.5cm}\n"
    "\n"
    "\n"
    "    \\huge\n"
    "    \\textsc{Report %s \\\\ \\vspace*{0.1cm}  %s}\n"
    "    \\vspace*{1cm}\n"
    "    \\end{center}\n"
    "\\end{titlepage}\n"
    )

tpl_1 = '\n'.join([
    '\n',
    '\\begin{sidewaysfigure}',
    '  \\centering',
    '      \\includegraphics[bb=50 50 554 770, height=\\textheight, angle=-90]{%s}',
    '  \\caption{%s}',
    '\\end{sidewaysfigure}',
    '\\clearpage',
    '%% to avoid "too many floats" error',
    '\n',
    ])

tpl_2 = '\n'.join([
    '\\begin{figure}',
    '    \\centering',
    '    \\includegraphics[bb=50 50 554 770, width=0.7\\textwidth, angle=-90]{%s}',
    '    \\includegraphics[bb=50 50 554 770, width=0.7\\textwidth, angle=-90]{%s}',
    '    \\caption{%s}',
    '\\end{figure}',
    ])

# 1,3,2,4, tl,bl,tr,br
tpl_4 = '\n'.join([
    '\\begin{sidewaysfigure}',
    '    \\begin{minipage}[c]{0.5\\textheight}',
    '        \\includegraphics[bb=50 50 554 770, height=0.42\\textheight, angle=-90]{%s}',
    '        \\includegraphics[bb=50 50 554 770, height=0.42\\textheight, angle=-90]{%s}',
    '    \\end{minipage}',
    '    \\begin{minipage}[c]{0.5\\textheight}',
    '        \\includegraphics[bb=50 50 554 770, height=0.42\\textheight, angle=-90]{%s}',
    '        \\includegraphics[bb=50 50 554 770, height=0.42\\textheight, angle=-90]{%s}',
    '    \\end{minipage}',
    '    \\caption{%s}',
    '\\end{sidewaysfigure}',
    '\\clearpage',
    '%% to avoid "too many floats" error',
    '\n',
    ])


def escape_latex_special_chars(input_string):
    """ escape latex special chars """
    import re
    # Define a dictionary of LaTeX special characters and their escaped forms
    latex_special_chars = {
        '\\': r'\textbackslash{}',  # Backslash
        '&': r'\&',
        '%': r'\%',
        '$': r'\$',
        '#': r'\#',
        '_': r'\_',
        '{': r'\{',
        '}': r'\}',
        '~': r'\textasciitilde{}',
        '^': r'\textasciicircum{}',
        '<': r'\textless{}',
        '>': r'\textgreater{}'
    }

    # Use regex to replace each special character with its escaped version
    escaped_string = re.sub(
        '|'.join(re.escape(key) for key in latex_special_chars.keys()),
        lambda match: latex_special_chars[match.group()],
        input_string
    )

    return escaped_string


def same_path(a, b):
    """ get path of file b assumed to be at the same path as file a,
    supports wildcast '*' in filename b
    """
    import os.path
    import glob
    p,f = os.path.split(a)
    # search by name, hopefully only ONE found
    fns = glob.glob(os.path.join(p, b))
    if len(fns) != 1:
        raise Exception("Cannot locate (or too many) heatflow .csv with '%s'" % heatflow_fname)
    return fns[0]

def get_csv_data(fname, columns=[0,1], factor=None):
    """ get data from csv file, can specify columns, also suuports
    factor for things like simple unit conversion.  factor should have
    the same length as the columns.
    """
    if factor is None:
        factor = [1.0] * len(columns)
    with open(fname, 'r') as f:
        data = []
        for line in f:
            if line:
                try:
                    sps = line.split(',')
                    d = [float(sps[i]) * m for i,m in zip(columns, factor)]
                    data.append(d)
                except ValueError:
                    # skip lines that can't be converted
                    continue
    return data

BPLOT_LS_ORDER = ['1','3','4','8','12','13','14','16','27','36','22','23','24','25','26','27','28','29',]

def mod_gpl(g, settings, skip_irrecoverable=False):
    """ this function parses the 'settings' dict used in entry, and apply
    actions to the gnuplot gpl object.

    Some action like 'reverse' (usually for injection) should only be done once.
    This can be done by setting the skip_irrecoverable = True. It's better to
    skip those actions when repeated modifying .gpl files.
    """
    from copy import deepcopy
    if 'replace_series' in settings:
        ls = BPLOT_LS_ORDER
        def replace_s(s, t, i):
            if 'data' in t:
                s.data = t['data']
                del t['data']
            if 'title' in t:
                s.title = t['title']
                del t['title']
            if 'with' in t:
                s.withstyle = t['with']
                del t['with']
            if 'ls' in t:
                s.set_ls(t['ls'])
                del t['ls']
            for k in t.keys():
                raise Exception("'Error in replace_series[%i]: key %s not recognised" % (i,k))
        for i,(s,t) in enumerate(zip(g.series, settings['replace_series'][:len(g.series)])):
            replace_s(s, t, i)
        for j,t in enumerate(settings['replace_series'][len(g.series):]):
            g.series.append(deepcopy(g.series[-1]))
            s = g.series[-1]
            s.set_with('linespoints')
            s.set_ls(ls[j+i])
            replace_s(s, t, j)

    if 'xlabel' in settings:
        if settings['xlabel'] is None:
            g.xlabel = ''
        else:
            g.xlabel = settings['xlabel']

    if 'ylabel' in settings:
        if settings['ylabel'] is None:
            g.ylabel = ''
        else:
            g.ylabel = settings['ylabel']

    if 'xrange' in settings:
        g.set('xrange', settings['xrange'])
        if settings['xrange'] is None:
            g.unset('xrange')

    if 'yrange' in settings:
        g.set('yrange', settings['yrange'])
        if settings['yrange'] is None:
            g.unset('yrange')

    if 'title' in settings:
        # replace if sepcified, other left alone
        g.set('title', settings['title'])

    if 'reverse' in settings and settings['reverse'] is True:
        if not skip_irrecoverable:
            def reverse_y(series):
                new_data = []
                for d in series.data:
                    new_data.append([d[0], -d[1]])
                series.data = new_data
                return series
            for s in g.series:
                s = reverse_y(s)

    if 'series_titles' in settings:
        for s,t in zip(g.series, settings['series_titles']):
            s.title = t

    if 'series_with' in settings:
        for s,t in zip(g.series, settings['series_with']):
            s.set_with(t)

    if 'series_ls' in settings:
        # NOTE bplot.exe generate ls sequence as:
        #      1 3 4 8 12 13 14 16 27 36 22 23 24 25 26 27 28 29
        for s,t in zip(g.series, settings['series_ls']):
            s.set_style(t)

    if 'apply_series_function' in settings:
        if not skip_irrecoverable:
            def apply_math(series, fn):
                new_data = []
                for d in series.data:
                    new_data.append(list(fn(d)))
                series.data = new_data
                return series
            for s in g.series:
                s = apply_math(s, settings['apply_series_function'])

    if 'apply_gpl_function' in settings:
        if not skip_irrecoverable:
            settings['apply_gpl_function'](g)

    if 'xrangemin' in settings:
        g.auto_xrange_min(settings['xrangemin'])

    if 'yrangemin' in settings:
        g.auto_yrange_min(settings['yrangemin'])

    if 'terminal' in settings:
        g.set('terminal', settings['terminal'])

    if 'add_cmd' in settings:
        for cmd in settings['add_cmd']:
            g._cmdlines.append(cmd + '\n')

    if 'xtics' in settings:
        g.xtics = settings['xtics']

    if 'ytics' in settings:
        g.ytics = settings['ytics']

    if 'mxtics' in settings:
        g.set('mxtics', settings['mxtics'])

    if 'mytics' in settings:
        g.set('mytics', settings['mytics'])

    if 'grid' in settings:
        g.set('grid', settings['grid'])

    return g

class FigureReport(object):
    def __init__(self, geometry, listings,
                 fig_path='./_plots/',
                 data_path='./field_data/',
                 title='',
                 starting_time=0.0,
                 report_name='report',
                 additional_t2listings=[],
                 ):
        self.geometry = geometry
        self.listings = listings
        self.title = title
        self.starting_time = starting_time

        self.bplot_lists = {}
        self.latex_lines = []
        self.gpl_mods = [] # !!! this is the real list of all figures, stored in each .gpl files

        # skip bplot, directly extract results and insert during self.mod_gpls
        self.additional_t2listings=additional_t2listings

        #  TODO-redo these:
        self._captions = {} # caption to be shown exported excel

        self.fig_path = fig_path # where bplot outputs .gpl .eps etc.
        self.data_path = data_path # fielddata dir for bplot
        self.report_name = report_name

        if not os.path.exists(self.fig_path):
            os.mkdir(self.fig_path)
        if not os.path.exists(self.data_path):
            os.mkdir(self.data_path)

    def _new_entry(self, p):
        """ return the gpl file name """
        def entry_string(p):
            lines = ['#####%2i' % len(p['geners'])]
            lines += p['geners']
            if 'datafiles' in p:
                lines += p['datafiles']
            lines += ['\n'] * 3
            s = '\n'.join(lines)
            return s
        if p['type'] not in self.bplot_lists:
            self.bplot_lists[p['type']] = []
        self.bplot_lists[p['type']].append(entry_string(p))
        fname = p['type']+ ('%i' % len(self.bplot_lists[p['type']]))
        self.gpl_mods.append((fname+'.gpl',p))
        return fname

    def add_titlepage(self, title, report_number, date):
        self.latex_lines.append(titlepage % (title, report_number, date))
        self.latex_lines.append('\n\n')

    def add_chapter(self, name, set_counter=None):
        if isinstance(set_counter, int):
            self.latex_lines.append('\\setcounter{chapter}{%i}' % (set_counter-1))
        self.latex_lines.append('\\chapter{%s}' % name)
        self.latex_lines.append('\n')

    def add_lines(self, lines):
        self.latex_lines += lines

    def add_tex_file(self, filename):
        with open(filename, 'r') as f:
            self.latex_lines += f.readlines()

    def add_table(self, contents, header=None):
        from tabulate import tabulate
        self.latex_lines += [
            r'',
            r'\begin{center}',
            str(tabulate(contents, header, tablefmt='latex')),
            r'\end{center}',
            r'',
            ]

    def add_page_1(self, p1):
        # need empty plot ".eps" to avoid latex waiting
        fname = '.eps'
        if p1 is not None:
            fname = self._new_entry(p1) + '.eps'
            self.latex_lines.append(tpl_1 % (fname, p1['caption']))
            self._captions[fname.replace('.eps', '.gpl')] = p1['caption']

    def add_page_2(self, p1, p2, caption):
        # need empty plot ".eps" to avoid latex waiting
        fn1, fn2 = '.eps', '.eps'
        if p1 is not None:
            fn1 = self._new_entry(p1) + '.eps'
        if p2 is not None:
            fn2 = self._new_entry(p2) + '.eps'
        self.latex_lines.append(tpl_2 % (fn1, fn2, caption))
        self._captions[fn1.replace('.eps', '.gpl')] = caption + ' (a)'
        self._captions[fn2.replace('.eps', '.gpl')] = caption + ' (b)'

    def add_page_4(self, p1, p2, p3, p4, caption):
        # need empty plot ".eps" to avoid latex waiting
        fn = ['.eps'] * 4
        for i,p in enumerate([p1, p2, p3, p4]):
            if p is not None:
                fn[i] = self._new_entry(p) + '.eps'
        self.latex_lines.append(tpl_4 % (fn[0], fn[2], fn[1], fn[3], caption))
        self._captions[fn[0].replace('.eps', '.gpl')] = caption + ' (a)'
        self._captions[fn[1].replace('.eps', '.gpl')] = caption + ' (b)'
        self._captions[fn[2].replace('.eps', '.gpl')] = caption + ' (c)'
        self._captions[fn[3].replace('.eps', '.gpl')] = caption + ' (d)'

    def _write_bplot_cfg(self):
        """ generate file option.cfg to be used by bplot.exe """
        lines = [
            '[GeometryFile] ! Geometry (MULGRAPH) file name',
            os.path.normpath(self.geometry),
            '![Tough2File] ! TOUGH2 data file name',
            '! not used here',
            '!-----',
            '[ListingFiles,%i] ! listing file names' % len(self.listings),
            '\n'.join([os.path.normpath(p) for p in self.listings]),
            '!-----',
            '[WellListFiles,%i] ! list of columns and real well data file names' % len(self.well_lists),
            '\n'.join([os.path.normpath(p) for p in self.well_lists]),
            '!-----',
            '[ReadDataDirectory] ! real data directory',
            os.path.normpath(self.data_path)+os.sep,
            '[PlotOutputDirectory] ! plots output directory',
            os.path.normpath(self.fig_path)+os.sep,
            '[FontSize] ! font size: 18 for Evi',
            '18',
            '[PlotMainTitle] ! title',
            self.title,
            '[NumOfSeriesPerPlot] ! maximum of series per plot - 1 ',
            '50',
            '[HistoryStartingTime] ! starting time for history matching',
            '%f' % self.starting_time,
            '\n\n\n',
            '! file generated by make_report.py',
            '\n\n\n',
        ]
        with open('option.cfg', 'w') as f:
            f.write('\n'.join(lines))

    def make_bplot(self, clean_files=True):
        # this has to wait until here, cannot do it in __init__
        flist = {
            'HistMtd': 'tpl_mass_rate.list',
            'HistEth': 'tpl_enthalpy.list',
            'HistStd': 'tpl_steam_rate.list',
            'HistPab': 'tpl_pressu.list',
            'HistPdr': 'tpl_pressu_drawdown.list',
            'HistTmp': 'tpl_temp.list',
        }
        self.well_lists = []
        for t,lines in self.bplot_lists.items():
            with open(flist[t], 'r') as ftpl:
                flistname = flist[t].replace('tpl_', '_')
                self.well_lists.append(flistname)
                with open(flistname, 'w') as flst:
                    flst.write(''.join(ftpl.readlines()))
                    flst.write('\n'.join(lines))
                    flst.write('[END]\n\n')
        self._write_bplot_cfg()
        if clean_files:
            all_files = os.listdir(self.fig_path)
            del_files = [os.path.join(self.fig_path, f) for f in all_files if f.lower()[-4:] in ['.eps', '.gpl', '.png']]
            for f in del_files:
                os.remove(f)
        with open('bplot.in', 'w') as f:
            f.write('option.cfg\n')
        subprocess.call('bplot < bplot.in', shell=True)

    def mod_gpls(self, settings_for_all={}, additional_gpl_cmds=[],
                 filetype='eps', skip_irrecoverable=False):
        """ additional settings that applies on ALL Report can be supplied here """


        orig = os.getcwd()
        os.chdir(self.fig_path)
        for fn,m in self.gpl_mods:
            # print fn, ':', m
            g = gpl.gplfile(fn)

            # additional t2listings, extract directly, add to series
            if m['type'] == 'HistPab':
                block = m['geners'][0]
                for lst in self.additional_t2listings:
                    try:
                        x, y = lst.history(('e', block, 'fluid_pressure'))

                        xx = (Unit.Q(x, "sec") + Unit.Q(self.starting_time, "year")).to('year')
                        yy = Unit.Q(y, "Pa").to("bar")

                        new_series = deepcopy(g.series[-1])
                        new_series .data = []
                        for xxx,yyy in zip(xx.magnitude, yy.magnitude):
                            new_series.data.append([xxx, yyy])
                        new_series.title = lst.filename
                        new_series.withstyle = 'linespoints ls 3'
                        g.series.append(new_series)

                        # breakpoint()

                    except KeyError:
                        print(f"@@@@@@@@@@@ Failed to get {lst._h5} {('e', block, 'pressure')}")
                        continue

            if m['type'] == 'HistTmp':
                block = m['geners'][0]
                for lst in self.additional_t2listings:
                    try:
                        x, y = lst.history(('e', block, 'fluid_temperature'))
                        xx = (Unit.Q(x, "sec") + Unit.Q(self.starting_time, "year")).to('year')
                        yy = Unit.Q(y, "degC").to("degC")

                        new_series = deepcopy(g.series[-1])
                        new_series .data = []
                        for xxx,yyy in zip(xx.magnitude, yy.magnitude):
                            new_series.data.append([xxx, yyy])
                        new_series.title = lst.filename
                        new_series.withstyle = 'linespoints ls 3'
                        g.series.append(new_series)

                        # breakpoint()

                    except KeyError:
                        print(f"@@@@@@@@@@@ Failed to get {lst._h5} {('e', block, 'pressure')}")
                        continue


            g.set_filetype(filetype)
            g.set_custom_linestyles(linewidth=2, pointsize=1.2)
            # additional_gpl_cmds
            g._cmdlines.append(" \n")
            for cmd in additional_gpl_cmds:
                if not cmd.endswith('\n'):
                    cmd += '\n'
                g._cmdlines.append(cmd)
            g._cmdlines.append(" \n")
            # settings for all
            # m.update(settings_for_all) # maybe this will render other ones useless?
            g = mod_gpl(g, settings_for_all, skip_irrecoverable=skip_irrecoverable)
            # settings this plot, comes last, to overwrite
            g = mod_gpl(g, m, skip_irrecoverable=skip_irrecoverable)
            g.remove_abs_path()
            g.write()
            subprocess.call(['gnuplot', fn])
        os.chdir(orig)

    def export_xls(self, output_dir='_export'):
        """ export all bplot data to an Excel spreadsheets

        NOTE xlwt does not supports newer format .xlsx which allows more than 256 columns
             xlsxwriter is very similar to xlwt
             TODO upgrade to something more modern, which may have rather different interface

            import xlwt
            wb = xlwt.Workbook()
            ws = wb.add_sheet('%i' % i)
            wb.save(os.path.join(output_dir, '%s.xls' % self.report_name))
        """
        import xlsxwriter
        wb = xlsxwriter.Workbook(os.path.join(output_dir, '%s.xlsx' % self.report_name))
        orig = os.getcwd()
        os.chdir(self.fig_path)
        try:
            for i,(fn,m) in enumerate(self.gpl_mods):
                g = gpl.gplfile(fn)
                ws = wb.add_worksheet('%i' % i)
                ws.write(0, 0, self._captions[fn] + ':' + g.title)
                for ii,s in enumerate(g.series):
                    ofsx = 1 + ii * 2 # X
                    ofsy = ofsx + 1   # Y
                    ws.write(ofsx, 0, s.title); ws.write(ofsy, 0, s.title)
                    ws.write(ofsx, 1, g.xlabel); ws.write(ofsy, 1, g.ylabel)
                    for jj,v in enumerate(s.data):
                        ws.write(ofsx, 2+jj, v[0])
                        ws.write(ofsy, 2+jj, v[1])
        except Exception as e:
            print('Error while exporting...')
            print('    ', fn)
            print('    ', self._captions[fn] + ':' + g.title)
            print('    ', s.title)
            raise e
        os.chdir(orig)
        if not os.path.isdir(output_dir):
            if os.path.exists(output_dir):
                raise Exception("FigureReport.export_xls() output_dir '%s' is possibly a file." % output_dir)
            else:
                os.mkdir(output_dir)
        wb.close()

    def make_report(self, output_dir='_report', clean_files=False):
        # compose latex file and compile, within self.fig_path
        with open('report_template.tex', 'r') as ftpl:
            with open(self.fig_path + self.report_name + '.tex', 'w') as ftex:
                ftex.write(''.join(ftpl.readlines()))
                ftex.write('\n'.join(self.latex_lines))
                ftex.write('\n\n\\end{document}\n\n\n')

        orig = os.getcwd()
        os.chdir(self.fig_path)
        subprocess.call(['latex', f"{self.report_name}.tex"])
        subprocess.call(['dvips', '-P', 'pdf', f"{self.report_name}.dvi"])
        subprocess.call([
            'gs', '-sPAPERSIZE=a4', '-dSAFER', '-dBATCH', '-dNOPAUSE',
            '-sDEVICE=pdfwrite',
            f'-sOutputFile={self.report_name}.pdf',
            '-c', 'save', 'pop',
            '-f', f'{self.report_name}.ps',
            ])

        # subprocess.call('"C:\\Program Files\\MiKTeX 2.9\\miktex\\bin\\x64\\dvips.exe" -P pdf "%s.dvi"' % self.report_name)
        # subprocess.call('"C:\\Program Files\\gs\\gs9.21\\bin\\gswin64c.exe" -sPAPERSIZE=a4 -dSAFER -dBATCH -dNOPAUSE -sDEVICE=pdfwrite -sOutputFile="%s.pdf" -c save pop -f "%s.ps"' % (self.report_name, self.report_name), shell=True)
        if clean_files:
            for ext in ['.aux', '.dvi', '.log', '.maf', '.mtc', '.mtc0', '.out', '.ps', '.tex']:
                os.remove(self.report_name + ext)
        os.chdir(orig)
        if not os.path.isdir(output_dir):
            if os.path.exists(output_dir):
                msg1 = 'Failed to copy report %s into specified output_dir.' % os.path.join(self.fig_path, self.report_name+'.pdf')
                msg2 = "FigureReport.make_report() output_dir '%s' is possibly a file." % output_dir
                raise Exception(msg1 + '\n' + msg2)
            else:
                os.mkdir(output_dir)
        shutil.copy2(os.path.join(self.fig_path, self.report_name+'.pdf'), output_dir)




""" Lazy man's easy conversion between date strings and date tuple/date objects.

It's actually very easy to code it directly, plus it's better to reduce
dependency.  This module provides very few additional functionality.  One being
able to detect if a list of strings/date tuples are given, and act accordingly.
Use these few lines of code instead if you are not too lazy:

from datetime import datetime
d = datetime.strptime(s, '%Y-%m-%d').date()

from datetime import date
s = str(date((1995,2,3)))
s = str(date(*tp))
# tp is your existing tuple, s will be string '1995-02-03'
"""
import unittest


def date_tuple_to_string(tp):
    """ Converts date integer tuple (y,m,d) into string 'yyyy-mm-dd'.  Input
    supports both a single date-tuple as well as a list of date-tuples. """
    from datetime import date
    single = False
    if len(tp) == 3:
        if all([isinstance(i, int) for i in tp]):
            single = True
    if single:
        return str(date(*tp))
    else:
        # assuming input is a list of date-tuples
        return [str(date(*d)) for d in tp]

def _s2d(s):
    """ Converts a date string of format 'yyyy-mm-dd' or 'dd/mm/yyyy' into a
    python datetime.date object. """
    from datetime import datetime
    if '/' in s:
        frmt = '%d/%m/%Y'
    elif '-' in s:
        frmt = '%Y-%m-%d'
    else:
        raise Exception("Date string format only support 'yyyy-mm-dd' or 'dd/mm/yyyy'")
    return datetime.strptime(s, frmt).date()

def date_string_to_tuple(s):
    """ Converts a date string of format 'yyyy-mm-dd' or 'dd/mm/yyyy' into a
    tuple of integers (y,m,d).  Input supports both a single string as well as a
    list of strings. """
    def d2t(d):
        return (d.year, d.month, d.day)
    if isinstance(s, list):
        return [d2t(_s2d(sd)) for sd in s]
    else:
        return d2t(_s2d(s))

def date_string_to_date(s):
    """ Converts a date string of format 'yyyy-mm-dd' or 'dd/mm/yyyy' into a
    python datetime.date object.  Input supports both a single string as well as
    a list of strings. """
    from datetime import date
    if isinstance(s, list):
        return [_s2d(sd) for sd in s]
    else:
        return _s2d(s)

def year_fraction_to_date(year_fraction):
    """ Converts a decimal/float year into a python datetime.date object.
    Input supports both a single float as well as a list of floats. """
    from datetime import date, timedelta
    def yf2d(yf):
        year = int(yf)
        fraction = yf - year
        # assuming 365.25 days per year
        day_of_year = int(fraction * 365.25)
        return date(year, 1, 1) + timedelta(days=day_of_year)
    if isinstance(year_fraction, list):
        return [yf2d(yf) for yf in year_fraction]
    else:
        return yf2d(year_fraction)

def year_fraction_to_date_str(year_fraction):
    """ Converts a decimal/float year into a string of format 'dd/mm/yyyy'.
    Input supports both a single float as well as a list of floats. """
    dd = year_fraction_to_date(year_fraction)
    if isinstance(dd, list):
        return [d.strftime("%d/%m/%Y") for d in dd]
    else:
        return dd.strftime("%d/%m/%Y")

def toYearFraction(date):
    """ converts python datetime objects into decimal/float years
    https://stackoverflow.com/a/6451892/2368167
    """
    from datetime import datetime as dt
    import time
    def sinceEpoch(date): # returns seconds since epoch
        return time.mktime(date.timetuple())
    s = sinceEpoch

    year = date.year
    startOfThisYear = dt(year=year, month=1, day=1)
    startOfNextYear = dt(year=year+1, month=1, day=1)

    yearElapsed = s(date) - s(startOfThisYear)
    yearDuration = s(startOfNextYear) - s(startOfThisYear)
    fraction = yearElapsed/yearDuration

    return date.year + fraction

def toYearFraction2(date):
    """ converts python datetime objects into decimal/float years
    https://stackoverflow.com/a/6451892/2368167

    fix Epoch issue?
    """
    from datetime import datetime as dt
    from datetime import date as dd
    import time
    def sinceEpoch(date): # returns seconds since epoch
        return date - dd(1970,1,1)
    s = sinceEpoch

    year = date.year
    startOfThisYear = dd(year=year, month=1, day=1)
    startOfNextYear = dd(year=year+1, month=1, day=1)

    yearElapsed = s(date) - s(startOfThisYear)
    yearDuration = s(startOfNextYear) - s(startOfThisYear)
    fraction = yearElapsed.total_seconds()/yearDuration.total_seconds()

    return date.year + fraction

def year_fraction(year, month, day):
    """ https://stackoverflow.com/a/36949905/2368167
    """
    import datetime
    date = datetime.date(year, month, day)
    start = datetime.date(date.year, 1, 1).toordinal()
    year_length = datetime.date(date.year+1, 1, 1).toordinal() - start
    return date.year + float(date.toordinal() - start) / year_length

class TestEasyDate(unittest.TestCase):
    """docstring for TestEasyDate"""
    def test_year_fraction(self):
        for d,f in [
            ((2023, 1, 1), 2023.0),
            ((2023, 6, 30), 2023.493),
            ((2023, 12, 31), 2023.997),
            ((2024, 1, 1), 2024.0),
        ]:
            result = year_fraction(*d)
            self.assertAlmostEqual(result, f, places=3)
                
    def test_year_fraction_mid_year(self):
        """Test year_fraction for middle of the year"""
        # Test July 1st (approximately mid-year)
        result = year_fraction(2023, 7, 1)
        self.assertGreater(result, 2023.4)
        self.assertLess(result, 2023.6)

    def test_year_fraction_to_date(self):
        """ test by starting with dd/mm/yyyy, convert to fraction, then back to dd/mm/yyyy, final one should be the same as the first one """
        ds = ['2011-03-05', '2011-3-6', '04/11/2011', '5/11/2011']
        for d in ds:
            dd = _s2d(d)
            f = toYearFraction2(dd)
            d2 = year_fraction_to_date(f)
            self.assertEqual(dd, d2)
        
        ds = ['04/11/2011', '05/01/2011']
        for d in ds:
            dd = _s2d(d)
            f = toYearFraction2(dd)
            d2 = year_fraction_to_date_str(f)
            self.assertEqual(d, d2)

    def test_tuple_to_string(self):
        s1 = (1995,10,2)
        s2 = [1995,10,2]
        ss = [s1, s2]
        for s in ss:
            self.assertEqual('1995-10-02', date_tuple_to_string(s))
        self.assertEqual(['1995-10-02']*2, date_tuple_to_string(ss))

    def test_string_to_tuple(self):
        d1 = '2011-03-05'
        d2 = '2011-3-5'
        d3 = '05/03/2011'
        d4 = '5/3/2011'
        ds = [d1,d2,d3,d4]
        for d in ds:
            self.assertEqual((2011,3,5), date_string_to_tuple(d))
        self.assertEqual([(2011,3,5)]*4, date_string_to_tuple(ds))

    def test_string_to_date(self):
        from datetime import date
        d1 = '2011-03-05'
        d2 = '2011-3-5'
        d3 = '05/03/2011'
        d4 = '5/3/2011'
        ds = [d1,d2,d3,d4]
        for d in ds:
            self.assertEqual(date(2011,3,5), date_string_to_date(d))
        self.assertEqual([date(2011,3,5)]*4, date_string_to_date(ds))



if __name__ == '__main__':
    unittest.main(verbosity=2)



import logging

from gimfu.config import *

from t2data import *
from t2incons import *

import os
from os import getcwd, sep, mkdir
from sys import argv
import json

SEC_IN_A_YEAR = 60.*60.*24.*365.25

def initLogger():
    """ initialise python built-in logging module for global use """
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    # to console
    con_formatter = logging.Formatter("-%(levelname)-10s %(message)s")
    con = logging.StreamHandler()
    con.setLevel(logging.INFO)
    con.setFormatter(con_formatter)
    logger.addHandler(con)
    # to log file
    logf_formatter = logging.Formatter("-%(levelname)-10s %(message)s\n"\
        "            %(filename)s:%(lineno)d")
    logf = logging.FileHandler('make_scenarios_2.log')
    logf.setLevel(logging.INFO)
    logf.setFormatter(logf_formatter)
    logger.addHandler(logf)
    logging.debug("logging initialised")
initLogger()

def make_save2incon(dirname=''):
    code = '\n'.join([
        "''' use: save2incon a.save b.incon [-reset_kcyc] '''",
        "from sys import argv",
        "from t2incons import t2incon",
        "if len(argv) < 2:",
        "    print('use: save2incon a.save b.incon [-reset_kcyc]')",
        "    exit(1)",
        "readFrom = argv[1]",
        "saveTo = argv[2]",
        "if len(argv) > 3:",
        "    opt = argv[3]",
        "else:",
        "    opt = ''",
        "inc = t2incon(readFrom)",
        "if opt == '-reset_kcyc':",
        "    inc.timing['kcyc'] = 1",
        "    inc.timing['iter'] = 1",
        "inc.write(saveTo)",
        "",
        ])
    with open(dirname+'save2incon.py','w') as f:
        f.write(code)

def make_sequential_batch_cmd(bases, dirname='', simulator='autough2_6f'):
    logging.info('Using simulator %s' % simulator)
    logging.info('Generating batch file: run_all_models.bat')
    logging.info('Generating batch file: make_long_listing.bat')
    fbat = open(dirname+'run_all_models.bat','w')
    fcombine = open(dirname+'make_long_listing.bat','w')
    fbat.write('date /t > time.log\n')
    fbat.write('time /t >> time.log\n')
    for i, b in enumerate(bases):
        f = open(dirname+b+'.in','w')
        f.write('\n'.join([
            b,
            b,
            b,
            '']))
        f.close()
        if i != 0:
            #fbat.write('cp '+ bases[i-1]+ '.save '+ bases[i]+ '.incon\n')
            fbat.write('python save2incon.py '+bases[i-1]+'.save '+bases[i]+'.incon -reset_kcyc\n')
        fbat.write('date /t >  '+ b +'.time\n')
        fbat.write('time /t >>  '+ b +'.time\n')
        fbat.write(('%s < ' % simulator)+ b + '.in\n')
        fbat.write('time /t >> '+ b +'.time\n')
        fbat.write('time /t >> time.log\n')
        if i == 0:
            fcombine.write('type '+ b + '.listing >  long.listing\n')
        else:
            fcombine.write('type '+ b + '.listing >> long.listing\n')

    # combine at the end of run anyway
    for i, b in enumerate(bases):
        if i == 0:
            fbat.write('type '+ b + '.listing >  long.listing\n')
        else:
            fbat.write('type '+ b + '.listing >> long.listing\n')

    fbat.close()
    fcombine.close()

def make_sequential_batch_cmd_sh(bases, dirname='', simulator='autough2_6f'):
    """
    Generate bash scripts for running all models and combining listings on Linux/macOS.
    """
    logging.info('Using simulator %s' % simulator)
    logging.info('Generating bash file: run_all_models.sh')
    logging.info('Generating bash file: make_long_listing.sh')
    fbat = open(os.path.join(dirname, 'run_all_models.sh'), 'w')
    fcombine = open(os.path.join(dirname, 'make_long_listing.sh'), 'w')
    fbat.write('#!/bin/bash\n')
    fbat.write('date > time.log\n')
    fbat.write('date +"%T" >> time.log\n')
    for i, b in enumerate(bases):
        with open(os.path.join(dirname, b + '.in'), 'w') as f:
            f.write('\n'.join([
                b,
                b,
                b,
                ''
            ]))
        if i != 0:
            fbat.write(f'python save2incon.py {bases[i-1]}.save {b}.incon -reset_kcyc\n')
        fbat.write(f'date > {b}.time\n')
        fbat.write(f'date +"%T" >> {b}.time\n')
        fbat.write(f'{simulator} < {b}.in\n')
        fbat.write(f'date +"%T" >> {b}.time\n')
        fbat.write(f'date +"%T" >> time.log\n')
        if i == 0:
            fcombine.write(f'cat {b}.listing > long.listing\n')
        else:
            fcombine.write(f'cat {b}.listing >> long.listing\n')

    # combine at the end of run anyway
    for i, b in enumerate(bases):
        if i == 0:
            fbat.write(f'cat {b}.listing > long.listing\n')
        else:
            fbat.write(f'cat {b}.listing >> long.listing\n')

    fbat.close()
    fcombine.close()
    # Make scripts executable
    os.chmod(os.path.join(dirname, 'run_all_models.sh'), 0o755)
    os.chmod(os.path.join(dirname, 'make_long_listing.sh'), 0o755)

def is_leap_year(y):
    if   int(y)%400 == 0: return True
    elif int(y)%100 == 0: return False
    elif int(y)%4   == 0: return True
    else:                 return False

def days_in_month(month, leap_year=False):
    if leap_year:
        d_month = [  31 , 29 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    else:
        d_month = [  31 , 28 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    return d_month[month - 1]

def date2str(d,m,y):
    ds, ms, ys = str(d), str(m), str(y)
    while len(ds) < 2: ds = '0'+ds
    while len(ms) < 2: ms = '0'+ms
    while len(ys) < 4: ys = '0'+ys
    return ds+'/'+ms+'/'+ys

def date2num_(enddate):

    d,m,y = enddate.split('/')
    months    = [ '01','02','03','04','05','06','07','08','09','10','11','12' ]
    d_month   = [  31 , 28 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    d_month_l = [  31 , 29 , 31 , 30 , 31 , 30 , 31 , 31 , 30 , 31 , 30 , 31  ]
    acum_ds   = [sum(d_month[:i]) for i in range(12)]
    acum_ds_l = [sum(d_month[:i]) for i in range(12)]
    ad_m   = dict(zip(months, acum_ds)) # accumulated days before this month
    ad_m_l = dict(zip(months, acum_ds_l)) # accumulated days before this month
    ds_m   = dict(zip(months, d_month)) # maximum days in this month
    ds_m_l = dict(zip(months, d_month_l)) # maximum days in this month
    # check and process d and m, this depends on data
    if m not in months:
        print(' Error, unable to convert ', enddate, ' to numeric format. check month.')
        sys.exit()
    if is_leap_year:
        if int(d) not in range(1,ds_m_l[m]+1):
            print(' Error, unable to convert ', enddate, ' to numeric format. check day.')
            sys.exit()
        num = float(y) + ((float(d) + float(ad_m_l[m]))/float(sum(d_month_l)))
    else:
        if int(d) not in range(1,ds_m[m]+1):
            print(' Error, unable to convert ', enddate, ' to numeric format. check day.')
            sys.exit()
        num = float(y) + ((float(d) + float(ad_m[m]))/float(sum(d_month)))
    return num

def date2num(date):
    try:
        date_value = float(date)
    except ValueError:
        date_value = date2num_(date)
    return date_value

def read_gener_file(fname):
    f = t2data_parser(fname,'r')
    line = f.readline()
    if line[:5] != 'GENER':
        # get rid of first line 'GENER' and check
        print('!!!!!!!!!! ERROR !!!!!!!!!!')
        print('Error reading GENER file: ', fname)
    gs = t2data()
    gs.read_generators(f)
    f.close()
    return gs.generatorlist

def lines_to_geners(lines,tmpfname):
    """ passing a list of strings and transform into a list of geners.
        the genres will alo be written into a file """
    # write
    lines_to_write = ['GENER'] + lines + ['']
    #lines.insert(0,'GENER')
    #lines.append('')
    ftmp = open(tmpfname,'w')
    ftmp.writelines('\n'.join(lines_to_write))
    ftmp.close()
    # read back
    return read_gener_file(tmpfname)

def lose_precision(num, frmt):
    """ Returns the float number that loses its precision after the
    formating.  This is done by putting numbers into string, using specified
    frmt and then back to float. """
    s = frmt % num
    return float(s)

def enable_t2data_time_higher_prec():
    """
    Temporarily modify PyTOUGH's t2data writing format specification, to allow
    tstart and tstop to be written with slightly higher precision: from '10.3e'
    to '10.4e'.  This can be dangerous, as negative time will be broken by
    '10.4e'.  Generally it is better to use save/incon file mechanism to control
    the much higher precision restart time, instead of playing with the spec.
    This however will not work if you want ot control higer precision of
    stopping time.
    """
    t2data_format_specification['param2'][1][0] = '10.4e' # tstart
    t2data_format_specification['param2'][1][1] = '10.4e' # tstop

# a list of file names that contains the building blocks (geners)
#
# For lines started with an '*', a config section will be read instead.
# eg.
#*TMAK_test
# will use section [*TMAK_test] as the source of GENERators
#
# Lines starting with '>', will read a number between '>' and '*'
# and repeat the *section the sepcified number of times
# eg.
#> 5 *ToRepeat
# will use *ToRepeat 5 times
#
# for lines startes with ':', are treated as direct lines from TOUGH2
# input file.
# eg.
#:hbn47hbn47                 -12     DMAK -1.000e-11 0.000e+00 2.000e+01 2.500e+05
#: 0.0000000e+00 8.4000000e+05 8.5000000e+05 9.0000000e+05
#: 1.0000000e+06 1.1000000e+06 1.2000000e+06 1.3000000e+06
#: 1.5000000e+06 1.9000000e+06 2.7000000e+06 2.8000000e+06
#: 3.0000000e+07 3.0000000e+07 1.8132376e+07 1.7823866e+07
#: 1.5017281e+07 9.7064135e+06 5.0435647e+06 4.1507355e+06
#: 3.8343393e+06 3.3400564e+06 3.3400564e+06 3.3400564e+06
# will cause the lines to be read as raw tough2 file (minus the first ':'
#
# For lines start with an '|', CFG file of making IMAK geners will be read after the
# second '|', and scaled (evenly) to the specified total injection.
# eg.
#|500.00|make_injection.cfg
# will run the make_IMAK_wells.py script and use make_injection.cfg to generate a
# list of GENERs to insert.
#
def collect_geners_from_cfg(cfg, cfg_entry_name):
    """ returns a list of generators and its spec, as a tuple of ([], {}).

    The spec is a optional record/info of the contents of the geners produced
    from this suroutine.
    """
    all_geners = []
    spec = {
        'external_geners_files': [], # external gener file names
        'well_stacks': [],
        'specifications': {},
    }
    if isinstance(cfg_entry_name, list):
        # a list (hopefully strings) is lines to be processed
        parts = cfg_entry_name
    else:
        # a single string is a config entry name
        parts = cfg.get_list(cfg_entry_name)
    i = 0
    t2verbatim = False
    while i < len(parts):
        line = parts[i]
        if line.strip()[0] == ':':
            if not t2verbatim:
                t2verbatim = True
                t2vname = 't2verbatim_'+str(i+1)+'.tmp'
                v_lines = []
            v_lines.append(line[line.find(':')+1:])
        elif t2verbatim:
            # write and read t2 verbatim
            print('  inserting verbatim: (', t2vname, ')')
            gs = lines_to_geners(v_lines,t2vname)
            all_geners = all_geners + gs
            # finishes verbatim
            t2verbatim = False
        if not t2verbatim:
            if line.strip()[0] == '*':
                # write and read a section as t2 gener
                s_lines = cfg.get_list(line.strip())
                print('  inserting section: ', line.strip())
                gs = lines_to_geners(s_lines,line[1:].strip()+'.tmp')
                all_geners = all_geners + gs
            elif line.strip()[0] == '>':
                # line must starts as >N* where N is number of repeating sections
                if line.find('*') == -1:
                    print('!!!!!!!!!! ERROR !!!!!!!!!!')
                    print('Error reading CFG line: ', line)
                    from sys import exit
                    exit()
                s_lines = cfg.get_list(line[line.find('*'):].strip())
                repeat = int(line[1:line.find('*')])
                print('  inserting section: ', line[line.find('*'):].strip(), ' by ', repeat, ' times ')
                gs = lines_to_geners(s_lines,line[line.find('*')+1:].strip()+'.tmp')
                for r in range(repeat):
                    all_geners = all_geners + gs
            elif line.strip()[0] == '|':
                inj_entry = line.strip()[1:].split('|')
                if len(inj_entry) != 2:
                    print('!!!!!!!!!! ERROR !!!!!!!!!!')
                    print('Error reading CFG line: ', line)
                    from sys import exit
                    exit()
                total_inj = float(inj_entry[0])
                inj_cfg = config(inj_entry[1])
                print('  inserting section: ', total_inj, ' kg/s of ', inj_entry[1].strip())
                from make_IMAK_wells import make_IMAK_wells_proc_cfg, make_IMAK_wells
                injpar = make_IMAK_wells_proc_cfg(inj_cfg)
                gs = make_IMAK_wells(injpar[0],injpar[1],injpar[2],injpar[3],injpar[4],injpar[5],injpar[6],total_inj)
                all_geners = all_geners + gs
            else:
                # read geners froma a specified file
                fname_geners = line.strip()
                print('  inserting GENER file: ' + fname_geners)
                gs = read_gener_file(fname_geners)
                all_geners = all_geners + gs
                spec['external_geners_files'].append(fname_geners)
                # optional spec
                fname_spec = fname_geners + '.spec.json'
                if os.path.isfile(fname_spec):
                    with open(fname_spec, 'r') as fspec:
                        print('    found associated geners spec: ' + fname_spec)
                        gs_spec = json.load(fspec)

                        well_stack_name = gs_spec['name']
                        spec['well_stacks'].append(well_stack_name)
                        spec['specifications'][well_stack_name] = gs_spec
                else:
                    spec['well_stacks'].append(None)
        i = i + 1
    if t2verbatim:
        # have to finish verbatim if it happends to be the last line
        print('  inserting verbatim: (', t2vname, ')')
        gs = lines_to_geners(v_lines,t2vname)
        all_geners = all_geners + gs
        # finishes verbatim
    return all_geners, spec

class GenerSection(object):
    """ A gener section is a list of generator objects that generally represents
    a part of a field, which can change over time.  It is specified as a list of
    enddates with a list of respective geners.  Geners is a list of t2data
    generator objects. """
    def __init__(self):
        super(GenerSection, self).__init__()
        self.enddates = []
        self.geners = []
        self.specs = []

    def add(self, enddate=None, gener_list=[], spec=None):
        """ If enddate is None, means no enddate, applies forever.  If
        gener_list is empty list [], means no geners until the enddate.  It does
        NOT make sense to have both empty date and empty list. """
        if enddate is None and gener_list == []:
            logging.error("GenerSection.add(): It does not make sense to have a section without both enddate and gener_list.")
            return
        self.enddates.append(enddate)
        self.geners.append(gener_list)
        self.specs.append(spec)

    def periodRange(self, i):
        """ returns the starting and end time of period of index i """
        e = self.enddates[i] # could raise IndexError
        if i - 1 < 0:
            s = None
        else:
            s = self.enddates[i-1]
        return (s,e)

    def inRange(self,t,range_or_i):
        """ return True if t is within range of the specified range (or period
        index)"""
        if isinstance(range_or_i,tuple):
            r = range_or_i
        else:
            r = self.periodRange(range_or_i)

        s_ok = False
        if r[0] is None:
            s_ok = True
        else:
            if r[0] < t:
                s_ok = True

        e_ok = False
        if r[1] is None:
            e_ok = True
        else:
            if r[1] >= t:
                e_ok = True

        return all([s_ok,e_ok])

    def indexAt(self, date):
        """ This gives the index of the time period that covers the date. """
        for i in range(len(self.enddates)):
            if self.inRange(date,i):
                return i
        return None

    def genersAt(self, date):
        """ This gives the geners of the time period that covers the date. """
        i = self.indexAt(date)
        if i is None:
            return []
        else:
            return self.geners[i]

    def specsAt(self, date):
        i = self.indexAt(date)
        if i is None:
            return {}
        else:
            return self.specs[i]

    def nextDateAt(self, date):
        """ Returns the next time-tag, if the date value is queal to the in-
        range period's tim-tag, the next tim-tag will be used.  If no next time-
        tag, return None. """
        i = self.indexAt(date)
        if i is None:
            return None
        ed = self.enddates[i]
        if ed == date:
            if i+1 < len(self.enddates):
                ed = self.enddates[i+1]
            else:
                ed = None
        return ed

    def read(self, cfg, s_name, offset_date=0.0):
        """ Reads from a cfg entry, which contains information about time
        varying specification of a part of field.  """
        if isinstance(s_name, str):
            sec_lines = cfg.get_list(s_name)
        elif isinstance(s_name,list):
            sec_lines = s_name
        else:
            logging.error("Unable to read %s" % s_name)

        lines = []
        for line in sec_lines:
            if line.startswith('@'):
                time = date2num(line[1:]) - offset_date # in simulation time
                gs, spec = collect_geners_from_cfg(cfg,lines)
                self.add(time, gs, spec)
                lines = [] # reset for next time period
            else:
                lines.append(line)
        if lines:
            # some left-over lines without a end time-tag
            gs, spec = collect_geners_from_cfg(cfg,lines)
            # print(spec)
            self.add(None, gs, spec)

        # print('~~~~~~~~~~', self.specs)

import unittest
class GenerSectionTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def _make_gener_sect(self,list_of_tuples):
        """ [ (1990.0, [x,y,z]), (1995.0, []) ] will make GenerSection obj of
        gs.enddates = [1990.0, 1995.0], gs.geners = [[x,y,z],[]]"""
        gs = GenerSection()
        gs.enddates, gs.geners = zip(*list_of_tuples)
        return gs

    def test_read(self):
        lines = [
        ':aaa11a   1                   0     IMAK        1.0',
        ':aaa22a   2                   0     IMAK        2.0',
        ':aaa33a   3                   0     IMAK        3.0',
        '@ 1000.0',
        ':bbb44b   4                   0     IMAK        4.0',
        ':bbb55b   5                   0     IMAK        5.0',
        ':bbb66b   6                   0     IMAK        6.0',
        '@ 2000.0',
        ]
        gs = GenerSection()
        gs.read(None,lines)
        print("%i time periods:" % len(gs.enddates))
        for d,gg in zip(gs.enddates, gs.geners):
            print("    %s @ (%s)" % (str(d),'), ('.join([str(g) for g in gg])))

        lines = [
        ':aaa11a   1                   0     IMAK        1.0',
        ':aaa22a   2                   0     IMAK        2.0',
        ':aaa33a   3                   0     IMAK        3.0',
        ]
        gs = GenerSection()
        gs.read(None,lines)
        print("%i time periods:" % len(gs.enddates))
        for d,gg in zip(gs.enddates, gs.geners):
            print("    %s @ (%s)" % (str(d),'), ('.join([str(g) for g in gg])))

        lines = [
        '@ 1000.0',
        ':bbb44b   4                   0     IMAK        4.0',
        ':bbb55b   5                   0     IMAK        5.0',
        ':bbb66b   6                   0     IMAK        6.0',
        ]
        gs = GenerSection()
        gs.read(None,lines)
        print("%i time periods:" % len(gs.enddates))
        for d,gg in zip(gs.enddates, gs.geners):
            print("    %s @ (%s)" % (str(d),'), ('.join([str(g) for g in gg])))

    def test_geners_at(self):
        gs = self._make_gener_sect([
            (1000.0, [1,2,3]),
            (2000.0, [4,5,6]),
            ])
        self.assertEqual(gs.genersAt( 989.0), [1,2,3])
        self.assertEqual(gs.genersAt(1000.0), [1,2,3])
        self.assertEqual(gs.genersAt(1500.0), [4,5,6])
        self.assertEqual(gs.genersAt(2000.0), [4,5,6])
        self.assertEqual(gs.genersAt(2000.1), [])

        gs = self._make_gener_sect([
            (None, [1,2,3]),
            ])
        self.assertEqual(gs.genersAt( 989.0), [1,2,3])
        self.assertEqual(gs.genersAt(1000.0), [1,2,3])
        self.assertEqual(gs.genersAt(1500.0), [1,2,3])
        self.assertEqual(gs.genersAt(2000.0), [1,2,3])
        self.assertEqual(gs.genersAt(2000.1), [1,2,3])

        gs = self._make_gener_sect([
            (1000.0, []),
            (None, [4,5,6]),
            ])
        self.assertEqual(gs.genersAt( 989.0), [])
        self.assertEqual(gs.genersAt(1000.0), [])
        self.assertEqual(gs.genersAt(1500.0), [4,5,6])
        self.assertEqual(gs.genersAt(2000.0), [4,5,6])
        self.assertEqual(gs.genersAt(2000.1), [4,5,6])

        gs = self._make_gener_sect([
            (1000.0, [1,2,3]),
            ])
        self.assertEqual(gs.genersAt( 989.0), [1,2,3])
        self.assertEqual(gs.genersAt(1000.0), [1,2,3])
        self.assertEqual(gs.genersAt(1500.0), [])
        self.assertEqual(gs.genersAt(2000.0), [])
        self.assertEqual(gs.genersAt(2000.1), [])

        gs = self._make_gener_sect([
            (1000.0, []),
            (2000.0, [4,5,6]),
            ])
        self.assertEqual(gs.genersAt( 989.0), [])
        self.assertEqual(gs.genersAt(1000.0), [])
        self.assertEqual(gs.genersAt(1500.0), [4,5,6])
        self.assertEqual(gs.genersAt(2000.0), [4,5,6])
        self.assertEqual(gs.genersAt(2000.1), [])

    def test_next_date_at(self):
        gs = self._make_gener_sect([
            (1000.0, [1,2,3]),
            (2000.0, [4,5,6]),
            ])
        self.assertEqual(gs.nextDateAt( 989.0), 1000.0)
        self.assertEqual(gs.nextDateAt(1000.0), 2000.0)
        self.assertEqual(gs.nextDateAt(1500.0), 2000.0)
        self.assertEqual(gs.nextDateAt(2000.0), None)
        self.assertEqual(gs.nextDateAt(2000.1), None)

        gs = self._make_gener_sect([
            (None, [1,2,3]),
            ])
        self.assertEqual(gs.nextDateAt( 989.0), None)
        self.assertEqual(gs.nextDateAt(1000.0), None)
        self.assertEqual(gs.nextDateAt(1500.0), None)
        self.assertEqual(gs.nextDateAt(2000.0), None)
        self.assertEqual(gs.nextDateAt(2000.1), None)

        gs = self._make_gener_sect([
            (1000.0, []),
            (None, [4,5,6]),
            ])
        self.assertEqual(gs.nextDateAt( 989.0), 1000.0)
        self.assertEqual(gs.nextDateAt(1000.0), None)
        self.assertEqual(gs.nextDateAt(1500.0), None)
        self.assertEqual(gs.nextDateAt(2000.0), None)
        self.assertEqual(gs.nextDateAt(2000.1), None)

        gs = self._make_gener_sect([
            (1000.0, [1,2,3]),
            ])
        self.assertEqual(gs.nextDateAt( 989.0), 1000.0)
        self.assertEqual(gs.nextDateAt(1000.0), None)
        self.assertEqual(gs.nextDateAt(1500.0), None)
        self.assertEqual(gs.nextDateAt(2000.0), None)
        self.assertEqual(gs.nextDateAt(2000.1), None)

        gs = self._make_gener_sect([
            (1000.0, []),
            (2000.0, [4,5,6]),
            ])
        self.assertEqual(gs.nextDateAt( 989.0), 1000.0)
        self.assertEqual(gs.nextDateAt(1000.0), 2000.0)
        self.assertEqual(gs.nextDateAt(1500.0), 2000.0)
        self.assertEqual(gs.nextDateAt(2000.0), None)
        self.assertEqual(gs.nextDateAt(2000.1), None)


    def tearDown(self):
        pass

def make_ver1(cfg):
    """ this is the original make_scenario, based on a list of specified time
    cuts, and each time cut is specified.  This is the older way of making
    scenarios, suitable for those have sorted out the scenarios nicely already.
    It may be harder to use than ver2 if start from scratch. """
    sname = cfg.get_value('ScenarioName').strip()
    #nsdat = t2data(cfg.get_value('NaturalStateDataFile').strip())
    prname = cfg.get_value('ProductionDataFile').strip()
    prdat = t2data(prname)
    prsav = t2incon(prname[:-3]+'save')
    dates = cfg.get_list('ScheduleDates')
    offset_date = cfg.get_value('HistoryStartingTime').strip()

    # processing info from CFG
    date_values = []
    try: offset_data_value = float(offset_date)
    except: offset_data_value = date2num(offset_date)
    for d in dates:
        try: date_values.append(float(d) - offset_data_value)
        except: date_values.append(date2num(d) - offset_data_value)

    namebase = os.path.basename(prname).split('_')[0].replace('pr','fut') + '_' + sname + '_'
    dirname = sname + '\\'
    sfilenames = []
    for i in range(1,len(dates)+1):
        sfilenames.append(namebase + '%02d'%i)
    make_sequential_batch_cmd(sfilenames,dirname)
    make_sequential_batch_cmd_sh(sfilenames,dirname)
    sfilenames = [dirname+s+'.dat' for s in sfilenames]
    prsav.write(dirname+sfilenames[0]+'.incon', reset=True)

    print('--------------------------------------------')
    print('Common building blocks:')
    geners_common, spec_common = collect_geners_from_cfg(cfg, 'Gener_Common')

    for i in range(len(dates)):
        print('--------------------------------------------')
        print('File ', i+1, ' enddate: ', dates[i], '(', date_values[i], ')')
        prdat.parameter['tstart'] = prdat.parameter['tstop']
        prdat.parameter['tstop'] = date_values[i] *60.0*60.0*24.0*365.25

        geners_this, spec_this = collect_geners_from_cfg(cfg, 'Gener_'+str(i+1))
        prdat.clear_generators()
        for g in geners_common: prdat.add_generator(g)
        for g in geners_this: prdat.add_generator(g)

        print('Writing scenario file: ', sfilenames[i], '...')
        prdat.write(sfilenames[i])

    print('--------------------------------------------')
    print('Done - Finished writing all scenario files.')

def make_ver2(cfg):
    """ This reads, process, and makes scenario using VERSION 2 INTERFACE, with
    time-varied ScenarioSections specified.  The config file must have
    [ScenarioSections] and [ScenarioEndingTime] specified.  Each section should
    have time-tags spliting section into parts,each applied to a time period.
    """
    scenario_spec = {}
    ##### collect misc. info
    sname = cfg.get_value('ScenarioName').strip()
    prname = cfg.get_value('ProductionDataFile').strip()
    prdat = t2data(prname)
    prsav = t2incon(prname[:-3]+'save')

    offset_date = date2num(cfg.get_value('HistoryStartingTime').strip())
    final_date = date2num(cfg.get_value('ScenarioEndingTime').strip())

    print_all_init_tables = False
    if cfg.check_optional('PrintAllInitialTables'):
        if cfg.get_value('PrintAllInitialTables').strip().lower() == 'true':
            print_all_init_tables = True

    print_conne_table = False
    if cfg.check_optional('PrintConnectionTables'):
        if cfg.get_value('PrintConnectionTables').strip().lower() == 'true':
            print_conne_table = True

    if cfg.check_optional('Simulator'):
        simulator = cfg.get_value('Simulator').strip()

    scenario_spec.update({
        'name': sname,
        'pr_model': prname,
        'date_offset': offset_date,
        'date_end': final_date,
        'simulator': simulator,
        'simulations': [],
        'well_stack_specification': {},
    })

    ##### compose sections for the scenario
    sections = []
    section_names = cfg.get_list('ScenarioSections')
    for s_name in section_names:
        sect = GenerSection()
        sect.read(cfg,s_name)
        sections.append(sect)

    ##### check all possible time period
    start_t = prdat.parameter['tstop'] / SEC_IN_A_YEAR + offset_date
    final_t = final_date
    logging.debug("before iterating s = %.2f f = %.2f" % (start_t,final_t))
    times = []
    while start_t < final_t:
        end_t = min([s.nextDateAt(start_t) for s in sections if s.nextDateAt(start_t) is not None] + [final_t])
        times.append((start_t,end_t))
        start_t = end_t

    ##### geneerating simulation file names
    namebase = os.path.basename(prname).split('_')[0].replace('pr','fut') + '_' + sname + '_'
    dirname = sname + sep
    try:
        mkdir(dirname)
    except OSError:
        pass
    sfilenames = [namebase + '%02d'%i for i in range(1,len(times)+1)]

    # the value will lose some precision as it's being written into file (via t2data)
    from functools import partial
    t2_prec = partial(lose_precision,frmt='%10.4e')
    # make t2data write input file with slightly higher prec tstart and tstop
    enable_t2data_time_higher_prec()

    ##### actually make these generators and input files
    actual_sfilenames, skipped = [], []
    cnt = 0
    import copy
    orig_output_times = copy.deepcopy(prdat.output_times)
    orig_timestep = prdat.parameter['timestep'][0]
    for (i,fname), (start_t,end_t) in zip(enumerate(sfilenames),times):
        logging.info("+++ %i from %.3f to %.3f" % (i+1,start_t,end_t))
        prdat.parameter['tstart'] = (start_t - offset_date) * SEC_IN_A_YEAR
        prdat.parameter['tstop'] = (end_t - offset_date) * SEC_IN_A_YEAR

        if print_conne_table:
            prdat.parameter['print_level'] = 3
        else:
            prdat.parameter['print_level'] = 1

        # avoid the negative time stepping in case stopping time is shorter than
        # the first initial time step
        prdat.parameter['timestep'][0] = orig_timestep
        true_tstart = t2_prec(prdat.parameter['tstart'])
        true_tstop = t2_prec(prdat.parameter['tstop'])
        t_diff = true_tstop - true_tstart
        if t_diff < prdat.parameter['timestep'][0]:
            prdat.parameter['timestep'][0] = t_diff

        # skip if the actual time diff is too small
        if t2_prec(t_diff) == 0.0:
            logging.warning("Sim %i skiped because time diff too small." % i)
            skipped.append("%.4f -- %.4f (%10.4e -- %10.4e)" % (start_t,end_t,true_tstart,true_tstop))
            continue

        # TODO: find a better rule to avoid negative time steping
        # don't need yearly output times if sim is very short.
        # if t_diff < 0.1 * SEC_IN_A_YEAR:
        #     prdat.output_times = {}
        # else:
        #     prdat.output_times = copy.deepcopy(orig_output_times)

        # avoid printing initial tables after second run
        cnt += 1
        if cnt > 1:
            if print_all_init_tables:
                prdat.parameter['option'][24] = 2
            else:
                prdat.parameter['option'][24] = 0

        # ensure simulation finishes - not restricting MCYC, need new autough2
        prdat.parameter['max_timesteps'] = -1

        gener_this = []
        specs_this = []
        for s in sections:
            # the period is actually FROM start_t + a_very_small_number TO end_t
            gener_this += s.genersAt(end_t)
            specs_this.append(s.specsAt(end_t))
            # print(specs_this)
        for g in gener_this:
            logging.debug("  ('%s','%s')" % (g.block,g.name))

        prdat.clear_generators()
        for g in gener_this:
            prdat.add_generator(g)

        logging.debug("Simulation file %s written." % fname)
        prdat.write(dirname+fname+'.dat')
        if i == 0:
            prsav.write(dirname+fname+'.incon')
        actual_sfilenames.append(fname)

        scenario_spec['simulations'].append({
            'filename': fname,
            'aut2_tstart_sec': true_tstart,
            'aut2_tstop_sec': true_tstop,
            'tstart_yr': start_t,
            'tstop_yr': end_t,
            'specs': specs_this,
        })

    ##### make helper batch file
    make_sequential_batch_cmd(actual_sfilenames, dirname, simulator=simulator)
    make_sequential_batch_cmd_sh(sfilenames,dirname, simulator=simulator)
    make_save2incon(dirname)

    ##### warn about any skips
    if len(skipped) > 0:
        logging.warning("%i simulations are skipped" % len(skipped))
        for skp in skipped:
            logging.warning(skp)

    logging.info('Generating scenario spec file: scenario_spec.json')
    with open(dirname + 'scenario_spec.json', 'w') as fspec:
        json.dump(scenario_spec, fspec, indent=4, sort_keys=True)

def main():
    # read CFG file, if not specified, default name is assumed
    cwd = getcwd() + sep
    cfg = config()
    if len(argv) > 1:
        cfg.read_from_file(cwd+argv[1])
    else:
        cfg.read_from_file(cwd+'make_scenarios.cfg')

    if cfg.check_optional('ScheduleDates'):
        logging.info("[ScheduleDates] found in config -> use version 1 make")
        logging.warning("Please note version 1 format make is legacy, which is harder to use, please use version 2 spec format for new projects. AY May 2014.")
        make_ver1(cfg)
    elif cfg.check_optional('ScenarioEndingTime') and cfg.check_optional('ScenarioSections'):
        logging.info("[ScenarioEndingTime] and [ScenarioSections] found in config -> use version 2 make")
        make_ver2(cfg)
    else:
        logging.error("Please make sure you are configuring the scenario either using version 1 or 2, but not both!")
    return 0

if __name__ == '__main__':
    main()



""" use: save2incon a.save b.incon [-reset_kcyc] """

from sys import *
from t2incons import *

if len(argv) < 2:
    print('use: save2incon a.save b.incon [-reset_kcyc]')

readFrom = argv[1]
saveTo = argv[2]

if len(argv) > 3:
    opt = argv[3]
else:
    opt = ''
inc = t2incon(readFrom)

if opt == '-reset_kcyc':
    inc.timing['kcyc'] = 1
    inc.timing['iter'] = 1

inc.write(saveTo)

